from __future__ import annotations, print_function
from JazzNote import JazzNote
from Utility import Utility
input = Utility.input
class Key:
    allFlats = ["C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B"]
    onlyNaturals = ["C", "D", "E", "F", "G", "A", "B"]
    flatAscii, sharpAscii = "b", "#"
    flat, sharp, doubleFlat, doubleSharp = "♭", "♯", "𝄫", "𝄪"
    useUnicodeChars = True
    useDoubleAccidentals = False

    def __add__(self, n:int) -> Key:
        return Key(Key.allFlats[(self.distanceFromC() + n) % 12])
    def __sub__(self, n:int) -> Key:
        return Key(Key.allFlats[(self.distanceFromC() - n) % 12])
    def __int__(self) -> int:
        return self.distanceFromC()
    def __init__(self, note: str, scaleFunction: JazzNote = None):
        if Key.isValid(note):
            self.note = (
                note.replace(Key.doubleSharp, Key.sharpAscii * 2)
                .replace(Key.doubleFlat, Key.flatAscii * 2)
                .replace(Key.sharp, Key.sharpAscii)
                .replace(Key.flat, Key.flatAscii)
            )
            if scaleFunction != None:
                self.note = self.onJazz(scaleFunction).note
                # input('asdf {} {} {}'.format(self,scaleFunction,self.onJazz(scaleFunction)))
        elif isinstance(note, Key):
            if scaleFunction == None:
                self.note = note.note
            else:
                self.note = self.onJazz(scaleFunction).note
        else:
            raise TypeError(
                "Key expects valid note in str or Key types, not {} {}".format(
                    type(note), note
                )
            )
        assert self.note

    def __eq__(self, other):
        return str(self) == str(other)

    def __repr__(self):
        return "Key('" + self.note + "')"

    def __str__(self, useUnicode=None, useDoubleAccidentals=None) -> str:
        if useUnicode == None:
            useUnicode = Key.useUnicodeChars
        if useUnicode == False:
            return self.note
        if useDoubleAccidentals == None:
            useDoubleAccidentals = Key.useDoubleAccidentals
        if useDoubleAccidentals:
            subs = (
                Key.flatAscii * 2,
                Key.sharpAscii * 2,
                Key.flatAscii,
                Key.sharpAscii,
            )
            replacements = (Key.doubleFlat, Key.doubleSharp, Key.flat, Key.sharp)
        else:
            subs = (Key.flatAscii, Key.sharpAscii)
            replacements = (Key.flat, Key.sharp)
        _str = self.note
        for s in range(len(subs)):
            sub, replacement = subs[s], replacements[s]
            _str = _str.replace(sub, replacement)
        if '#' in _str or 'b' in _str:
            raise ValueError('shitshit')
        return _str

    def __len__(self):
        return len(self.note)
    def __getitem__(self,i):
        return self.note[i]
    

    def inAllFlats(self):
        return Key(Key.allFlats[self.distanceFromC() % 12])

    @classmethod
    def accidentalsStr(cls, step: int, useAscii=True):
        if step < 0:
            return Key.flatAscii * abs(step)
        elif step > 0:
            return Key.sharpAscii * step
        else:
            return ""

    def distanceFromC(self) -> int:
        return Key.allFlats.index(self.note[0]) + JazzNote.accidentalsToDistance(
            self.note[1:]
        )

    def getAccidentalSum(self) -> int:
        return self.note.count(Key.sharpAscii) - self.note.count(Key.flatAscii)

    def removeAccidentals(self) -> Key:
        return Key(self.note[0])

    def onJazz(self, jazz: JazzNote) -> Key:
        jazz = JazzNote(jazz)
        scaleDegree = int(jazz.scaleDegree())

        newKeyDegree = Key(Key.onlyNaturals[
                               (Key.onlyNaturals.index(self.note[0]) \
                                + scaleDegree - 1) % len(Key.onlyNaturals)])

        rootDist = (newKeyDegree.distanceFromC() - Key(self.note).distanceFromC()) % 12
        newKeyAccidentals = self.getAccidentalSum() + jazz.getAccidentalSum()
        newKeyAccidentals = Key.accidentalsStr(newKeyAccidentals)
        '''print('self {} jazz {} scaledegree {} newkeydegree {} newKeyAccidentals {} rootdist {} == {}'.format(
            self, jazz, scaleDegree, newKeyDegree, newKeyAccidentals, rootDist, Key(str(newKeyDegree) + newKeyAccidentals)
        ))'''
        val = Key(str(newKeyDegree) + newKeyAccidentals)

        return Key(str(newKeyDegree) + newKeyAccidentals)
        return Key(JazzNote(jazz).byWay(self.getASCII()))

    def isBlack(self) -> bool:
        return self.distanceFromC() in (1, 3, 6, 8, 10)

    def isWhite(self) -> bool:
        return not self.distanceFromC() in (1, 3, 6, 8, 10)

    def getEnharmonic(self, keyDegreeTranspose: int):
        root = Key(
            Key.onlyNaturals[
                (Key.onlyNaturals.index(self.note[0]) + keyDegreeTranspose)
                % len(Key.onlyNaturals)
            ]
        )
        accSum = (
            self.getAccidentalSum()
            + self.removeAccidentals().distanceFromC()
            - root.distanceFromC()
        )
        if accSum > 6:
            accSum = -(12 - accSum)
        elif accSum < -6:
            accSum = 12 - abs(accSum)
        return Key(str(root) + Key.accidentalsStr(accSum))
        input("bloop {} {} {} {} ".format(self, keyDegreeTranspose, root, accSum))

    def getAllowedJazz(
        self,
        jazz: JazzNote,
        maxAccidentals=2,
    ) -> [JazzNote]:
        allowedJazz = []
        accSum = self.getAccidentalSum()
        enharmonic = self
        for i in range(7):
            if abs(enharmonic.getAccidentalSum()) <= maxAccidentals:
                allowedJazz.append(enharmonic)
            enharmonic = enharmonic.getEnharmonic(-1)
        """if accSum == 0:
            allowedJazz.append(enharmonic)"""
        # input('fdafdsa {} accSum{} maxAccs{} allowedJazz {}'.format(
        #    self,accSum,maxAccidentals,allowedJazz))
        return allowedJazz

    def getASCII(self):
        return self.note.replace(Key.flat, Key.flatAscii).replace(
            Key.sharp, Key.sharpAscii
        )

    @classmethod
    def isValid(self, note: str):
        if not (type(note) is str):
            # raise TypeError(noteName,' is not str.')
            return False
        if not (note[0] in Key.onlyNaturals):
            # raise ValueError('noteName[0] is not in noteNameNaturals, it is: ',noteName[0])
            return False
        sharps, flats = False, False  # (False,)*2
        sharpChars = (Key.sharp, Key.sharpAscii, Key.doubleSharp)
        flatChars = (Key.flat, Key.flatAscii, Key.doubleFlat)
        if len(note) > 1:
            if note[1] in sharpChars:
                sharps = True
            elif note[1] in flatChars:
                flats = True
            else:
                # raise ValueError(noteName[1]," is not 'b' or '#'")
                return False
            for char in note[1:]:
                if sharps:
                    if not char in sharpChars:
                        raise ValueError("sharps turned into ", char)
                elif flats:
                    if not char in flatChars:
                        raise ValueError("flats turned into ", char)
        return True



'''from classFiles.JazzNote import JazzNote
from classFiles.Change import Change
from classFiles.Utility import Utility'''
from Utility import Utility
from JazzNote import JazzNote
from Change import Change
from Graphics import Graphics
from tqdm import tqdm
from pprint import pprint


#from pprint import pprint
import random, math
input = Utility.input
print = Utility.print
change = Change(['1','5'])
#print(change.byWays(['Word']))
notWorkingWays = []
def testWays():
    ret = {}
    result = ''
    notWorkingWays = {}
    for i in range(1361,1362):
        print('in testWays, i =',i)
        change = Change.makeFromChangeNumber(i)

        #input(i, change, change.byWays('Carnatic'))
        ways = [way for way in Change.validWays if "Poem" not in way]
        #random.seed(50)
        #random.shuffle(ways)

        for w,way in tqdm(enumerate(ways)):


            print('top of way loop')
            _result = ''
            try:
                #print(w, way, end='----->   ')
                _result += '{} {} -> '.format(w, way)
                changeByWay = change.byWays([way])
                if type(changeByWay) in (Change, list, tuple):
                    changeByWay = ', '.join([str(v) for v in changeByWay])
                _result += '{}\n'.format(changeByWay)
                #print(changeByWay,end='')
            except Exception as e:
                _result = ''
                _result += '\t# {} {} \t\t\t{}\n'.format(w,way,e)
                #print('\t#',w,way,end=', ')
                notWorkingWays[way] = w
                #print('\t!!!!Not Yet Working!!!!')
            result += _result
            #print()
            print('doing',w,way)

    non_working_result = '\nNon working ways num = {}\n {}\n'.format(len(notWorkingWays),', '.join( [ str(i) for i in notWorkingWays]))
    result = non_working_result + result + non_working_result

    ret['result'] = result
    ret['notWorkingWays'] = notWorkingWays

    try:
        change:Change
        ret['nextBadWay'] = change.byWays([list(notWorkingWays.keys())[0]])

    except IndexError:
        ret['nextBadWay'] = None
    pprint(ret)

    return ret

if __name__ == '__main__':
    print('passing testWays()...')
    pprint(testWays())
    print('done passing testWays()')
    print('done executing succesfully')


from __future__ import annotations, print_function
import pyx, math, os
from JazzNote import JazzNote
from Key import Key
from Change import Change
from Colour import Colour
from Utility import Utility

class Fretboard:
    height = 256
    lineWidth = round(height / 64)
    # marginSize measured in str space size
    marginSize = 0.5
    octaveWidthProportion = 2
    octaveWidth = height * 3.5

    
    lineColour = (0, 0, 0, 255)
    lineColourInvert = (255, 255, 255, 255)
    indicationColour = (255, 255, 255, 255)
    indicationColourInvert = (255, 255, 255, 255)
    indicationLineColour = (0, 0, 0, 255)  # (128,128,128,255)
    indicationLineColourInvert = (0, 0, 0, 255)  # (128,128,128,255)
    canvasBGColour = (255, 255, 255, 0)

    showFrets = 12
    # instruments default to Guitar values if not specified.
    instruments = {
        "Guitar": {
            "strNotes": ["E", "A", "D", "G", "B", "E"],
            "fretIndicators": (5, 7, 9, 15, 17, 19),
            "fretIndicatorsDouble": [12, 24],
            "showFrets": 24,
            "doubleDotsOffset": 9,
        },
        "DADGAD": {"strNotes": ["D", "A", "D", "G", "A", "D"]},
        "DADGBD": {"strNotes": ["D", "A", "D", "G", "B", "D"]},
        "Mando": {
            "strNotes": ["G", "D", "A", "E"],
            "fretIndicators": (5, 7, 10, 15, 17, 19),
            "fretIndicatorsDouble": [12],
        },
        "Uke": {
            "strNotes": ["G", "C", "E", "A"],
            "fretIndicators": (5, 7, 10, 12, 15),
            "fretIndicatorsDouble": [],
            "showFrets": 19,
        },
    }

    def __init__(self, instrument="Guitar", showFrets=None):
        # input('{} {}'.format(Fretboard.instruments.keys(),instrument))
        if instrument in Fretboard.instruments.keys():
            self.instrument = instrument
        else:
            raise KeyError(
                "you need to add this instrument: "
                + str(instrument)
                + ", to Fretboard.instruments"
            )
        if "strNotes" in Fretboard.instruments[instrument].keys():
            self.strNotes = Fretboard.instruments[instrument]["strNotes"]
        else:
            raise ValueError("you need to add strNotes to this instrument")
        for n in self.strNotes:
            if JazzNote.isAlphabetNoteStr(n):
                pass
            else:
                raise TypeError("strNotes expects notes in alphabet format Ab")

        if showFrets is None:
            if "showFrets" in Fretboard.instruments[instrument].keys():
                self.showFrets = Fretboard.instruments[instrument]["showFrets"]
            else:
                self.showFrets = Fretboard.instruments["Guitar"]["showFrets"]

        if "fretIndicators" in Fretboard.instruments[instrument].keys():
            self.fretIndicators = Fretboard.instruments[instrument]["fretIndicators"]
        else:
            self.fretIndicators = Fretboard.instruments["Guitar"]["fretIndicators"]

        if "fretIndicatorsDouble" in Fretboard.instruments[instrument].keys():
            self.fretIndicatorsDouble = Fretboard.instruments[instrument][
                "fretIndicatorsDouble"
            ]
        else:
            self.fretIndicators = Fretboard.instruments["Guitar"][
                "fretIndicatorsDouble"
            ]

        if "doubleDotsOffset" in Fretboard.instruments[instrument].keys():
            self.doubleDotsOffset = Fretboard.instruments[instrument][
                "doubleDotsOffset"
            ]
        else:
            self.doubleDotsOffset = Fretboard.instruments["Guitar"]["doubleDotsOffset"]

    def getFretboardPath(
        self,
        change: Change,
        key: str,
        greyScale=False,
        invertColour=False,
        returnName=True,
        returnDirectory=True,
        includeGraphicsPath=True,
        pathSeperator="/",
        resolution: int = None,
        fretBoardType: str = None,
        filetype: str = "png",
        externalGraphicsPath=False,
        includeExtension=True,
    ) -> str:
        key = JazzNote.noteNameFlats[JazzNote.distanceFromC(key)]
        if resolution is None:
            resolution = Fretboard.height
        _str = ""
        from Graphics import Graphics
        if returnDirectory:
            if includeGraphicsPath:
                # raise TypeError('includeGraphicsPath path should be false right now')
                if greyScale:
                    if not invertColour:
                        _str += Graphics.fretboardsBWDirectory
                    else:
                        _str += Graphics.fretboardsBWInvDirectory
                else:
                    if not invertColour:
                        _str += Graphics.fretboardsDirectory
                    else:
                        _str += Graphics.fretboardsInvDirectory
            else:
                _str += self.instrument + "/"

            if externalGraphicsPath:
                _str = _str.replace(Graphics.directory, Graphics.directoryExternal)

            _str = os.path.join(_str,str(resolution) + "px_" + filetype + "/",key )

        if returnName:
            _str = os.path.join(_str,"{}_{}".format(
                self.instrument,
                change.getChangeNumber(decorateChapter=False, addOneToBookPage=True),
            ))
            if includeExtension:
                _str += "." + filetype
        _str = _str.replace("\\", pathSeperator)
        _str = _str.replace("/", pathSeperator)
        _str = _str.replace("Fretboard", self.instrument)

        return _str

    def renderFretboards(
        self,
        changes: [Change],
        keys: [str],
        greyScale,
        invertColour=False,
        indicateNotesWithFCircle=False,
        fretsAreEquidistant=False,
        connectFretboardLines=True,
        showImage=False,
        makeColourSchemes=True,
        makeNewIndicatorDiagrams=False,
        useRadiusLines = True,
        filetypes=["pdf"],
        resolution=64,
    ):
        color = pyx.color
        canvas = pyx.canvas
        trafo = pyx.trafo
        style = pyx.style
        path = pyx.path
        deco = pyx.deco
        _noteColours = Colour.getTransposedColours(neutralColours=True)
        # Each fret space is .94387 as wide as the previous fret space
        if type(keys) == list and all(
            [JazzNote.isAlphabetNoteStr(key) for key in keys]
        ):
            pass
        else:
            raise TypeError(
                "keys expects a list of alphabet keys. not {} with type == {}".format(
                    keys, type(keys)
                )
            )

        if fretsAreEquidistant:
            _fretWidths = [
                self.octaveWidth / self.showFrets for i in range(self.showFrets + 1)
            ]
            _firstFretWidth = _fretWidths[0]
        else:
            _firstFretWidth = (
                self.octaveWidth / self.showFrets
            ) * self.octaveWidthProportion ** (-6 / 12)
            _fretWidths = [
                _firstFretWidth * self.octaveWidthProportion ** ((12 - i) / 12)
                for i in range(self.showFrets + 1)
            ]
        _fretX = _fretWidths[0]
        _strVertSpace = self.height / (len(self.strNotes) + self.marginSize * 2 - 1)
        _marginSpace = _strVertSpace * self.marginSize
        _imWidth = sum(_fretWidths[:-1]) + _fretX + _marginSpace
        usingRaster = "png" in filetypes
        from Graphics import Graphics
        from FCircle import FCircle
        usingVector = any([i in Graphics.vectorFiletypes for i in filetypes])
        vectorFCircleResolution = FCircle.vectorSaveResolutions[0]
        _FCirclesMadePng = []
        print(
            "beginning renderFretboards() with these {} change(s) {}\n in the key(s): {}\nfretBoard.instrument: {}".format(
                len(changes),
                ", ".join([str(c.getChangeNumber()) for c in changes]),
                keys,
                self.instrument,
            )
        )
        for k in keys:
            assert JazzNote.isAlphabetNoteStr(k), "shit"
            _colourSchemes = []
            for g in (False, True):
                for i in (False, True):
                    if (i == invertColour and g == greyScale) or makeColourSchemes:
                        print("stats i: {} g: {}".format(i, g))
                        _colourSchemes.append(
                            {
                                "greyScale": g,
                                "invertedColours": i,
                                                            }
                        )
                        if indicateNotesWithFCircle:
                            #raise TypeError ('shit')
                            _colourSchemes[-1]["canvasFCircles"] = []
                            _colourSchemes[-1]["canvasFCirclesInverted"] = []
                            _fCircleFiletypes = [] if not usingRaster else ["png"]
                            _whichCache = ('canvasFCircles',
                            greyScale, invertColour, rootColourKey, resolutionresolution, _fCircleFiletypes)
                            try:
                                Fretboard.cache
                            except NameError:
                                Fretboard.cache = {}

                            try:
                                _colourSchemes[-1]["canvasFCircles"] = Fretboard.cache[_whichCache]
                            except NameError:
                                print("key is {}!!!".format(k))
                                for st in range(12):
                                    _colourSchemes[-1]["canvasFCircles"].append(
                                        Graphics.renderFCircle(
                                            change=Change.makeFromSet([st]),
                                            greyScale=g,
                                            invertColour=i,
                                            rootColourKey=k,
                                            resolution=vectorFCircleResolution,
                                            saveOnlyOneKey=True,
                                            returnAsPyXCanvas=True,
                                            filetypes= _fCircleFiletypes,
                                        )
                                    )
                                    _colourSchemes[-1]["canvasFCirclesInverted"].append(
                                        Graphics.renderFCircle(
                                            change=Change.makeFromSet([st]),
                                            greyScale=g,
                                            invertColour=not i,
                                            rootColourKey=k,
                                            resolution=vectorFCircleResolution,
                                            saveOnlyOneKey=True,
                                            returnAsPyXCanvas=True,
                                            filetypes=_fCircleFiletypes,
                                        )
                                    )
                                    # print('come here again',len(_colourSchemes[-1]['canvasFCirclesInverted'][-1]))
                                    # input(str(_colourSchemes[-1]['canvasFCirclesInverted'][-1].bbox()))
                                Fretboard.cache[_whichCache] = _colourSchemes[-1]["canvasFCircles"]
                        if i == False:
                            _colourSchemes[-1]["lineColour"] = pyx.color.rgb(
                                Fretboard.lineColour[0] / 255,
                                Fretboard.lineColour[1] / 255,
                                Fretboard.lineColour[2] / 255,
                            )
                            _colourSchemes[-1]["indicationLineColour"] = pyx.color.rgb(
                                Fretboard.indicationLineColour[0] / 255,
                                Fretboard.indicationLineColour[1] / 255,
                                Fretboard.indicationLineColour[2] / 255,
                            )
                            _colourSchemes[-1]["indicationColour"] = pyx.color.rgb(
                                Fretboard.indicationColour[0] / 255,
                                Fretboard.indicationColour[1] / 255,
                                Fretboard.indicationColour[2] / 255,
                            )
                        else:
                            _colourSchemes[-1]["lineColour"] = pyx.color.rgb(
                                Fretboard.lineColourInvert[0] / 255,
                                Fretboard.lineColourInvert[1] / 255,
                                Fretboard.lineColourInvert[2] / 255,
                            )
                            _colourSchemes[-1]["indicationLineColour"] = pyx.color.rgb(
                                Fretboard.indicationLineColourInvert[0] / 255,
                                Fretboard.indicationLineColourInvert[1] / 255,
                                Fretboard.indicationLineColourInvert[2] / 255,
                            )
                            _colourSchemes[-1]["indicationColour"] = pyx.color.rgb(
                                Fretboard.indicationColourInvert[0] / 255,
                                Fretboard.indicationColourInvert[1] / 255,
                                Fretboard.indicationColourInvert[2] / 255,
                            )

            for change in changes:
                _fretX = _fretWidths[0]
                for col in range(len(_colourSchemes)):
                    _colourSchemes[col]["canvas"] = canvas.canvas()

                _noteNames = [
                    JazzNote.noteNameFlats[(i + JazzNote.distanceFromC(k)) % 12]
                    for i in change.getSemitones()
                ]
                _outerFretboardBounds = {}
                if usingRaster:
                    _im = Image.new(
                        "RGBA",
                        (round(_imWidth), round(self.height)),
                        Fretboard.canvasBGColour,
                    )
                    _draw = ImageDraw.Draw(_im)
                if usingVector:
                    _canvas = canvas.canvas()
                    _trafo = trafo.scale(sx=4 / _imWidth, sy=4 / -_imWidth, x=0, y=0)
                    _lineWidthVec = style.linewidth(
                        Fretboard.lineWidth / Fretboard.height
                    )
                    # _trafo = trafo.scale(sx=1 , sy=-1 , x=0, y=0)
                    _outerFretboardPath = path.path()
                _stringY = self.height - _marginSpace
                _stringNotes = []
                _stringYs = []
                _fretXs = []
                for s, string in enumerate(self.strNotes):
                    if usingRaster:
                        _draw.line(
                            (_fretX, _stringY, _imWidth - _marginSpace, _stringY),
                            fill=Fretboard.lineColour,
                            width=Fretboard.lineWidth,
                        )
                    if usingVector:
                        # _canvas.insert()
                        if not connectFretboardLines or s not in (
                            0,
                            len(self.strNotes) - 1,
                        ):
                            print(
                                "string at {} {} {} {}".format(
                                    _fretX, _stringY, _imWidth - _marginSpace, _stringY
                                )
                            )

                            # _linePath = path.path(path.moveto(_fretX,_stringY),
                            #                      path.lineto(_imWidth-_marginSpace,_stringY),
                            #                      path.closepath())

                            # _canvas.stroke(_linePath,
                            #               [style.linewidth.THICK,_trafo,color.rgb(255/255,255/255,255/255)])
                            for c in _colourSchemes:
                                c["canvas"].stroke(
                                    path.line(
                                        _fretX,
                                        _stringY,
                                        _imWidth - _marginSpace,
                                        _stringY,
                                    ),
                                    [_lineWidthVec, c["lineColour"], _trafo],
                                )
                        elif s == 0:
                            _outerFretboardBounds["L"] = _fretX
                            _outerFretboardBounds["R"] = _imWidth - _marginSpace
                            _outerFretboardBounds["T"] = _stringY
                        elif s == len(self.strNotes) - 1:
                            _outerFretboardBounds["B"] = _stringY
                    _stringYs.append(_stringY)
                    _stringY -= _strVertSpace
                    _stringNotes.append([])
                    for fret in range(self.showFrets + 1):
                        _stringNotes[-1].append(
                            JazzNote.noteNameFlats[
                                (JazzNote.distanceFromC(string) + fret) % 12
                            ]
                        )

                # Populate fretXs
                for fret in range(self.showFrets + 1):
                    _fretXs.append(_fretX)
                    # Draw fret

                    if usingRaster:
                        _draw.line(
                            (_fretX, _stringYs[0], _fretX, _stringYs[-1]),
                            fill=Fretboard.lineColour,
                            width=Fretboard.lineWidth,
                        )
                    if usingVector:
                        if not connectFretboardLines or s not in (
                            0,
                            self.showFrets - 1,
                        ):
                            for c in _colourSchemes:
                                c["canvas"].stroke(
                                    path.line(
                                        _fretX, _stringYs[0], _fretX, _stringYs[-1]
                                    ),
                                    [_trafo, _lineWidthVec, c["lineColour"]],
                                )
                        else:
                            pass
                    # Draw fret marker
                    _mW = _marginSpace / 5
                    _mX = _fretX - _fretWidths[fret] / 2
                    _mY = (_stringYs[0] + _stringYs[-1]) / 2
                    _mYoffset = _mW * self.doubleDotsOffset
                    # input('hasdf {}'.format(self.fretIndicatorsDouble))
                    # Draw fret indicators
                    if (fret) in self.fretIndicators:
                        if usingRaster:
                            _draw.ellipse(
                                [_mX - _mW, _mY - _mW, _mX + _mW, _mY + _mW],
                                fill=(0, 0, 0, 255),
                            )
                        if usingVector:
                            for c in _colourSchemes:
                                c["canvas"].draw(
                                    path.circle(_mX, _mY, _mW),
                                    [
                                        _lineWidthVec,
                                        _trafo,
                                        deco.stroked([c["indicationLineColour"]]),
                                        deco.filled([c["indicationLineColour"]]),
                                    ],
                                )

                    elif (fret) in self.fretIndicatorsDouble:
                        if usingRaster:
                            _draw.ellipse(
                                [
                                    _mX - _mW,
                                    _mY - _mW - _mYoffset,
                                    _mX + _mW,
                                    _mY + _mW - _mYoffset,
                                ],
                                fill=(0, 0, 0, 255),
                            )
                            _draw.ellipse(
                                [
                                    _mX - _mW,
                                    _mY - _mW + _mYoffset,
                                    _mX + _mW,
                                    _mY + _mW + _mYoffset,
                                ],
                                fill=(0, 0, 0, 255),
                            )

                        if usingVector:
                            for c in _colourSchemes:
                                c["canvas"].draw(
                                    path.circle(_mX, _mY + _mYoffset, _mW),
                                    [
                                        _lineWidthVec,
                                        _trafo,
                                        deco.stroked([c["indicationLineColour"]]),
                                        deco.filled([c["indicationColour"]]),
                                    ],
                                )
                                c["canvas"].draw(
                                    path.circle(_mX, _mY - _mYoffset, _mW),
                                    [
                                        _lineWidthVec,
                                        _trafo,
                                        deco.stroked([c["indicationLineColour"]]),
                                        deco.filled([c["indicationColour"]]),
                                    ],
                                )
                    _fretX += _fretWidths[fret]
                if connectFretboardLines and usingVector:
                    _outerFretboardPath = path.path()
                    _outerFretboardPath.append(
                        path.moveto(
                            _outerFretboardBounds["L"], _outerFretboardBounds["T"]
                        )
                    )
                    _outerFretboardPath.append(
                        path.lineto(
                            _outerFretboardBounds["R"], _outerFretboardBounds["T"]
                        )
                    )
                    _outerFretboardPath.append(
                        path.lineto(
                            _outerFretboardBounds["R"], _outerFretboardBounds["B"]
                        )
                    )
                    _outerFretboardPath.append(
                        path.lineto(
                            _outerFretboardBounds["L"], _outerFretboardBounds["B"]
                        )
                    )
                    _outerFretboardPath.append(path.closepath())
                    for c in _colourSchemes:
                        c["canvas"].stroke(
                            _outerFretboardPath,
                            [_lineWidthVec, c["lineColour"], _trafo],
                        )

                vectorFCirclesMade = []

                for s, string in enumerate(self.strNotes):
                    for fret in range(self.showFrets + 1):
                        if _stringNotes[s][fret] in _noteNames:
                            if len(_fretXs) > fret + 1:
                                if _fretXs[fret + 1] - _fretXs[fret] > _marginSpace * 2:
                                    _L = _fretXs[fret] - _marginSpace * 2
                                    _T = _stringYs[s] - _marginSpace
                                    _R = _fretXs[fret]
                                    _B = _stringYs[s] + _marginSpace
                                    # print('a width', _R - _L)
                                else:
                                    _R = _fretXs[fret]
                                    # _L = _fretXs[max(0, fret - 1)]
                                    _L = _fretXs[max(0, fret - 1)]
                                    _L = max(_L, _R - _marginSpace * 2)
                                    _T = _stringYs[s] - (_R - _L) / 2
                                    _B = _stringYs[s] + (_R - _L) / 2
                                    # print('b width', _R - _L)
                            else:
                                _R = _fretXs[fret]
                                _L = _fretXs[max(0, fret - 1)]
                                _L = max(_L, _R - _marginSpace * 2)
                                _T = _stringYs[s] - (_R - _L) / 2
                                _B = _stringYs[s] + (_R - _L) / 2
                                # print('width', _R - _L)
                            if fret == 0:
                                pass
                                # _L += _marginSpace * 2
                                # _R += _marginSpace * 2
                            if not indicateNotesWithFCircle:
                                if usingRaster:
                                    _draw.ellipse(
                                        (_L, _T, _R, _B),
                                        fill=(0, 0, 0, 255),
                                        outline=Fretboard.lineColour,
                                    )
                                if usingVector:
                                    
                                    for c in _colourSchemes:
                                        # Draw invisible indicator to keep the frame consistent
                                        if s == 0 and fret == 0 \
                                                and all([Key(k, scaleFunction=note) not in self.strNotes for note in change]):
                                            _L = _fretXs[fret] - _marginSpace * 2

                                        _noteKey = Key(_stringNotes[s][fret])
                                        _noteColour = _noteColours[_noteKey.distanceFromC()]
                                        #print(_noteKey.distanceFromC(),Key(k).distanceFromC())
                                        #input(_noteKey.distanceFromC()-Key(k).distanceFromC())
                                        _noteChange = Change(noteset=[(_noteKey.distanceFromC() - Key(k).distanceFromC())%12])
                                        #input(_noteChange)
                                        c['indicationColour']  = color.rgb(_noteColour[0]/255,_noteColour[1]/255,_noteColour[2]/255)
                                            
                                            
                                        c["canvas"].draw(
                                            path.circle(
                                                (_R + _L) / 2,
                                                (_T + _B) / 2,
                                                (_R - _L) / 2,
                                            ),
                                            [
                                                _lineWidthVec,
                                                _trafo,
                                                deco.stroked(
                                                    [c["indicationLineColour"]]
                                                ),
                                                deco.filled([c["indicationColour"]]),
                                            ],
                                        )
                                        _rad = abs(_R - ((_R + _L) / 2))
                                        if useRadiusLines:
                                            c['canvas'].draw(path.line(
                                                (_R + _L) / 2,
                                                (_T + _B) / 2,
                                                (_R + _L) / 2 + _rad * math.cos(_noteChange.getAngles(returnDegrees=False,goQuarterTurnAnticlockwise=True)[0]),
                                                (_T + _B) / 2 + _rad * math.sin(_noteChange.getAngles(returnDegrees=False,goQuarterTurnAnticlockwise=True)[0]),
                                            ),
                                            [_lineWidthVec,deco.stroked([c["indicationLineColour"]]),
                                            _trafo,c["indicationLineColour"]])
                            elif indicateNotesWithFCircle:

                                _iWidth = _R - _L
                                if True or s == 0 and fret == 0:
                                    _maxWidth = min(
                                        abs(_stringYs[1] - _stringYs[0]),
                                        abs(_fretXs[1] - _fretXs[0]),
                                    )
                                # input( 'an important message:\nself.showFrets: {} _R: {} _L {} iWidth {} fretXs {}'.format(
                                #    self.showFrets,_R,_L,_iWidth,_fretXs))

                                _FCircleColourKey = k  # JazzNote.noteNameFlats[(JazzNote.noteNameFlats.index(_note) )%12]# + _startingDistFromC

                                _changeToIndicate = Change.makeFromSet(
                                    [
                                        (
                                            JazzNote.distanceFromC(
                                                _stringNotes[s][fret]
                                            )
                                            - JazzNote.distanceFromC(k)
                                        )
                                        % 12
                                    ]
                                )
                                _FCirclePathRaster = Graphics.getFCirclePath(
                                    _changeToIndicate,
                                    _FCircleColourKey,
                                    int(_iWidth),
                                    includeFilename=True,
                                    greyScale=greyScale,
                                    invertColour=(fret == 0),
                                )
                                # input('_FCirclePathRaster = {} \nisfile? {}'.format(os.path.normpath(_FCirclePathRaster),os.path.isfile(_FCirclePathRaster)))
                                if (
                                    not os.path.isfile(_FCirclePathRaster)
                                    and "png" in filetypes
                                ):  # FCircle does not exist
                                    if _iWidth == 0:
                                        print(
                                            "Unfortunately _iWidth is 0 for some reason so skipping renderFCircle"
                                        )
                                    else:
                                        if usingRaster:

                                            # input('happened')
                                            if (
                                                _changeToIndicate,
                                                int(_iWidth),
                                            ) not in _FCirclesMadePng:
                                                Graphics.renderFCircle(
                                                    change=_changeToIndicate,
                                                    resolution=int(_iWidth),
                                                    rootColourKey="C",
                                                    saveOnlyOneKey=False,
                                                    filetypes=["png"],
                                                )
                                                print(
                                                    "did make it because we had not made {} yet.".format(
                                                        (
                                                            _changeToIndicate,
                                                            int(_iWidth),
                                                        )
                                                    )
                                                )
                                                _FCirclesMadePng.append(
                                                    (_changeToIndicate, int(_iWidth))
                                                )
                                elif _iWidth != 0 and (not makeNewIndicatorDiagrams):
                                    pass
                                if _iWidth != 0:  # Stupid bug
                                    if usingRaster:
                                        _FCircleFileRaster = Image.open(
                                            _FCirclePathRaster.replace("/", "\\")
                                        ).convert("RGBA")
                                        _im.paste(
                                            _FCircleFileRaster,
                                            (int(_L), int(_T)),
                                            _FCircleFileRaster,
                                        )
                                    if usingVector:
                                        for colourScheme in _colourSchemes:
                                            """_FCirclePathVec = Graphics.getFCirclePath(
                                                _changeToIndicate,
                                                _FCircleColourKey,
                                                resolution=FCircle.vectorSaveResolutions[0], includeFilename=True,
                                                greyScale=colourScheme['greyScale'],
                                                invertColour=(fret == 0) if colourScheme['invertedColours'] == False else fret!=0,
                                                filetype='svg'
                                            )"""
                                            _canvasFCircle = colourScheme[
                                                "canvasFCircles"
                                                + ("Inverted" if fret == 0 else "")
                                            ][_changeToIndicate[0].semitonesFromOne()]
                                            # input('{}'.format(_iWidth,'fhgjfghjfghj'))
                                            _FCircleYScale = -2 * (_iWidth / _maxWidth)
                                            _FCircleXScale = 2 * (_iWidth / _maxWidth)
                                            _FCircleYScale = -2 * (
                                                min(_maxWidth, _iWidth)
                                                / vectorFCircleResolution
                                            )
                                            _FCircleXScale = 2 * (
                                                min(_maxWidth, _iWidth)
                                                / vectorFCircleResolution
                                            )

                                            # input('xscale {} yscale {} iWidth {} maxWidth.... {} vectorFCircleResolution {}'.format(
                                            #    _FCircleXScale,_FCircleYScale,_iWidth,_maxWidth,vectorFCircleResolution))

                                            _FCircleScaleTrafo = trafo.scale(
                                                _FCircleXScale, _FCircleYScale, 0, 0
                                            )

                                            _canvasFCircleScaled = canvas.canvas()
                                            _canvasFCircleScaled.insert(
                                                _canvasFCircle, [_FCircleScaleTrafo]
                                            )

                                            _FCircleTranslateTrafo = trafo.translate(
                                                _L, _T
                                            )

                                            colourScheme["canvas"].insert(
                                                _canvasFCircleScaled,
                                                [_FCircleTranslateTrafo, _trafo],
                                            )

                                            # input(_FCirclePathVec)
                                            # c['canvas'].insert(epsfile.epsfile(0, 0, _FCirclePathVec))
                                            # c['canvas'].insert(svgfile.svgfile(0, 0, _FCirclePathVec,parsed=True,width=self.height/8))

                if usingRaster:
                    _fullnameRaster = self.getFretboardPath(
                        change,
                        k,
                        greyScale,
                        invertColour=invertColour,
                        filetype="png",
                        resolution=resolution,
                    )
                    _pathnameRaster = self.getFretboardPath(
                        change,
                        k,
                        greyScale,
                        invertColour=invertColour,
                        filetype="png",
                        returnName=False,
                        resolution=resolution,
                    )
                    print(
                        "saving raster file:///{} to file:///{}".format(
                            _fullnameRaster.replace("\\", "/"),
                            _pathnameRaster.replace("\\", "/"),
                        )
                    )

                if usingVector:
                    bounds = path.path()
                    boundsL = _fretXs[0] - _marginSpace
                    boundsR = _fretXs[-1]
                    boundsT = _stringYs[0] + _marginSpace
                    boundsB = _stringYs[-1] - _marginSpace
                    bounds.append(path.moveto(boundsL, boundsT))
                    bounds.append(path.lineto(boundsR, boundsT))
                    bounds.append(path.lineto(boundsR, boundsB))
                    bounds.append(path.lineto(boundsL, boundsB))
                    # bounds.append(path.lineto(boundsL, boundsT))
                    bounds.append(path.closepath())
                    # color.transparency(.5)
                    # input(bounds)
                    for c in _colourSchemes:
                        boundsCanvas = canvas.canvas()
                        boundsCanvas.stroke(
                            bounds, [color.rgb(0, 1, 0), color.transparency(1), _trafo]
                        )
                        c["canvas"].insert(boundsCanvas)
                    for filetype in [
                        f for f in filetypes if f in Graphics.vectorFiletypes
                    ]:
                        # input('scheeeeemz ({}) \n{} \nfiletypes {}'.format(len(_colourSchemes),_colourSchemes,filetypes))

                        assert filetype != "eps"
                        assert len(_colourSchemes) == 4
                        for c in _colourSchemes:
                            _fullname = self.getFretboardPath(
                                change,
                                k,
                                greyScale=c["greyScale"],
                                invertColour=c["invertedColours"],
                                filetype=filetype,
                                resolution=resolution,
                            )

                            _pathname = self.getFretboardPath(
                                change,
                                k,
                                greyScale=c["greyScale"],
                                invertColour=c["invertedColours"],
                                returnName=False,
                                filetype=filetype,
                                resolution=resolution,
                            )
                            assert "eps" not in _fullname
                            print(
                                "saving vector file:///{} to file:///{}".format(
                                    _fullname, _pathname.replace("/", "/")
                                )
                            )
                            if not os.path.isdir(_pathname):
                                Utility.makeDirectory(_pathname)

                            c["canvas"].writetofile(_fullname)
                            # boundsCanvas.writetofile(_fullname)
                    # input('asdas {}'.format(len(_colourSchemes)))
                if showImage and usingRaster:
                    _im.show()
                    input("showing imagge")
                if makeColourSchemes:
                    if usingRaster:
                        Graphics.transposeColourSchemes(
                            _im,
                            _fullnameRaster,
                            k,
                        )
                    from Project import Project
                    if Project.findLinesWhereBookGetItemOccurred:
                        input(
                            "lines where getItem heppened {}".format(
                                Book.linesWhereGetItemHappened
                            )
                        )
                if "png" in filetypes:
                    if not os.path.isdir(_pathnameRaster):
                        Utility.makeDirectory(_pathnameRaster)
                    _im.save(_fullnameRaster)

                makeNewIndicatorDiagrams = False
                # print('finished with fretboard {} {} {}'.format(change, key,change.getChangeNumber(addOneToBookPage=True)))
            # del _colourSchemes


import os, random
from wayofchange import Project
import Utility
input = Utility.input
class Syllabic:
    # Espanol not finished needs to be sorted alphabetically
    validLanguages = ["Syllabic", "Shakespeare", "Espanol"]
    language = "Shakespeare"

    wordTypes = {
        "N": "Noun",
        "p": "Plural",
        "h": "Noun phrase",
        "V": "Verb (usually participle)",
        "t": "Transitive verb",
        "i": "Intransitive verb",
        "A": "Adjective",
        "v": "Adverb ",
        "C": "Conjunction",
        "P": "Preposition ",
        "!": "Interjection",
        "r": "Pronoun ",
        "D": "Definite article",
        "I": "Indefinite article",
        "o": "Nominative",
    }
    print("started going through moby words...")
    if True:  # or language == "Syllabic":
        with open(
                os.path.join(Project.directory, "mobypos.txt"), encoding="utf-8"
        ) as f:
            # with open("espanol.txt") as f:
            # with open('espanol.txt', encoding="ISO-8859-1") as f:
            # with open("espanol.txt") as f:
            mobyWords = {}
            mobyWordsByFirstLetter = {}
            for line in f:
                _word = line[: line.index("\\")]
                _wordType = line[line.index("\\"): line.index("\\") + 2]
                _wordType = _wordType.replace("\\", "")
                # mobyWords.append(line.strip())
                mobyWords[_word] = {"partOfSpeech": _wordType}
                _firstLetter = _word[0].lower()

                if not _firstLetter in mobyWordsByFirstLetter:
                    mobyWordsByFirstLetter[_firstLetter] = {}
                mobyWordsByFirstLetter[_firstLetter][_word] = {
                    "partOfSpeech": _wordType
                }
                # print(mobyWordsByFirstLetter)
    if True:
        # elif language == 'Shakespeare':
        with open(os.path.join(Project.directory, "shakespeare.txt")) as f:
            shakespeareWords = set
            shakespeareWordsByLetter = {}
            censorChars = [
                '"',
                ".",
                ";",
                ":",
                ",",
                "!",
                "'",
                "-",
                "?",
                ".",
                "*",
                "(",
                ")",
                ">",
                "<",
                "[",
                "]",
            ]
            omitWordWithCharsInMiddle = ["|", ".", "@", "_"]
            _wordsClean = []

            for l, line in enumerate(f):
                line = line.strip()
                _words = line.split(" ")
                for w, word in enumerate(_words):
                    if len(word) > 0:
                        _word = word[:]
                        while len(_word) > 0 and _word[-1] in censorChars:
                            _word = _word[:-1]
                        while len(_word) > 0 and _word[0] in censorChars:
                            _word = _word[1:]
                    if len(_word) > 1:
                        if _word.isupper():
                            _word = _word.lower()
                        # Filter numbers and weird computer strings
                        if len(_word) > 2 and any(
                                [char in ["|", ".", "@"] for char in _word[1:-1]]
                        ):
                            pass
                        elif any([c.isdigit() for c in _word]):
                            pass
                        elif any([c in ["|", ".", "@", "_"] for c in _word]):
                            _wordsClean.append(_word)
                            _wordsClean[-1] = (
                                _wordsClean[-1]
                                .replace("|", " ")
                                .replace("@", " ")
                                .replace(".", " ")
                                .replace("_", " ")
                            )
                        if _word[0].lower() in shakespeareWordsByLetter:
                            shakespeareWordsByLetter[_word[0].lower()].append(
                                _word
                            )
                        else:
                            shakespeareWordsByLetter[_word[0].lower()] = []
                # input('line {} {}'.format(l,' '.join(_wordsClean)))
            _wordsClean = set(_wordsClean)
            for letter in shakespeareWordsByLetter:
                shakespeareWordsByLetter[letter] = list(
                    set(shakespeareWordsByLetter[letter])
                )

    if language == "Espanol":

        # TODO# basic drunk version not tewsted
        espanolWords = []
        with open("espanol.txt", encoding="ISO-8859-1") as f:
            for l, line in enumerate(f):
                espanolWords.append(line.strip())
        raise ValueError(
            "Not done this yet.. maybe needs alphabetisation not sure"
        )

    # input('mobyWords created {} mobyWords created'.format('\n'.join(mobyWords)))
    # print(mobyWords.keys(),random.randint(1,1000))
    @classmethod
    def getRandomEnglishWord(cls, firstLetter=None, useShakespeare=None) -> str:
        if useShakespeare is None:
            useShakespeare = Syllabic.language == "Shakespeare"
        if firstLetter is None:
            _word = list(Syllabic.mobyWords.keys())[
                random.randint(0, len(Syllabic.mobyWords) - 1)
            ]
        else:
            firstLetter = firstLetter.lower()
            if not useShakespeare:
                _wordList = Syllabic.mobyWordsByFirstLetter
                _word = list(_wordList[firstLetter].keys())[
                    random.randint(0, len(_wordList[firstLetter]) - 1)
                ]
            else:
                _wordList = Syllabic.shakespeareWordsByLetter
                _word = list(_wordList[firstLetter])[
                    random.randint(0, len(_wordList[firstLetter]) - 1)
                ]

        return _word


from JazzNote import JazzNote
from Change import Change
from Fretboard import Fretboard
from Utility import Utility
from Graphics import Graphics
input = Utility.input
print = Utility.print
import math, PIL, os
from PIL import ImageDraw, Image
class Accordion:
    height = 256
    standard120buttons = True
    buttonsAcross = 20
    buttonMarginX = 0.618 / 2
    buttonMarginY = buttonMarginX
    canvasBGColour = (255, 255, 255, 0)
    buttonBGColour = (255, 255, 255, 255)
    buttonIndicatedBGColour = (0, 0, 0, 255)
    buttonOutlineColour = (0, 0, 0, 255)
    buttonIndicatedOutlineColour = (255, 255, 255, 255)
    rowOffsetByButtons = 0.333
    indicateNotes = ["C", "Eb", "Gb", "A"]
    indicateRow = 1
    if standard120buttons:
        buttonRows = [["Db"]]
        buttonRows = [
            [["Db"]],
            [["A"]],
            [["A", "C#", "E"]],
            [["A", "C", "E"]],
            [["A", "C#", "G"]],
            [["A", "C", "F#"]],
        ]

        for r, row in enumerate(buttonRows):
            # input('row {}'.format(row))
            for b in range(buttonsAcross):
                _firstButton = row[0]
                # input('b {} firstbutton {}'.format(b,_firstButton))
                if b != 0:
                    buttonRows[r].append([])
                    for note in _firstButton:
                        buttonRows[r][-1].append(
                            JazzNote.noteNameFlats[
                                (JazzNote.distanceFromC(note) + 7 * b) % 12
                            ]
                        )

        for row in buttonRows:
            print(row)
            print()
        print("generated accordion notes")

    @classmethod
    def getAccordionPath(
        cls,
        change: Change,
        key: str,
        greyScale,
        invertColour,
        returnName=True,
        returnDirectory=True,
        pathSeperator="/",
        resolution: int = None,
        filetype: str = "png",
        includeGraphicsDirectory=True,
        externalGraphicsPath=False,
        includeFileExtension=True,
    ) -> str:
        from Graphics import Graphics
        key = JazzNote.noteNameFlats[JazzNote.distanceFromC(key)]
        if resolution is None:
            resolution = Fretboard.height
        _str = ""
        if returnDirectory:
            if greyScale:
                if invertColour:
                    _str += Graphics.accordionsBWInvDirectory
                else:
                    _str += Graphics.accordionsBWDirectory
            else:
                if invertColour:
                    _str += Graphics.accordionsInvDirectory
                else:
                    _str += Graphics.accordionsDirectory
            if includeGraphicsDirectory:
                _str = os.path.join(_str,Graphics.getFiletypeResolutionFolderName(
                    filetype=filetype, resolution=resolution
                ))
                _str += "/" + key + "/"
                if externalGraphicsPath:
                    _str = _str.replace(Graphics.directory, Graphics.directoryExternal)
            else:
                _str = (
                    "Accordion/"
                    + Graphics.getFiletypeResolutionFolderName(
                        filetype=filetype, resolution=resolution
                    )
                    + "/"
                    + key
                    + "/"
                )
        if returnName:
            _str += "Accordion_{}".format(
                change.getChangeNumber(decorateChapter=False, addOneToBookPage=True),
            )
            if includeFileExtension:
                _str += "." + filetype
        _str = _str.replace("\\", pathSeperator)
        _str = _str.replace("/", pathSeperator)
        return _str

    @classmethod
    def renderAccordion(
        cls,
        change: Change,
        key: str,
        greyScale,
        invertColour,
        indicateNotesWithCircle=True,
        indicateNotesWithFCircle=True,
        showIm=False,
    ):
        Image = PIL.Image

        #raise TypeError("asdfasdf")
        _buttonHeight = 2 * Accordion.height / 12 / (1 + Accordion.buttonMarginY)
        _buttonWidth = _buttonHeight
        _imWidth = (_buttonWidth * (1 + Accordion.buttonMarginX)) * len(
            Accordion.buttonRows[0]
        )  # (Accordion.rowOffsetByButtons) * r
        _imWidth += (
            _buttonWidth
            * (Accordion.rowOffsetByButtons)
            * (len(Accordion.buttonRows) - 1)
        )
        # input('{}'.format(len(Accordion.buttonRows)))
        _im = Image.new(
            "RGBA", (round(_imWidth), round(Accordion.height)), Accordion.canvasBGColour
        )
        _draw = ImageDraw.Draw(_im)
        _buttonY = Accordion.buttonMarginY * _buttonHeight / 2
        _changeNotes = [
            JazzNote.noteNameFlats[JazzNote.distanceFromC(i)]
            for i in change.byWays(key)
        ]
        # input('pirateship {}'.format(_changeNotes))
        for r, row in enumerate(Accordion.buttonRows):
            _buttonX = (
                _buttonWidth * (Accordion.rowOffsetByButtons) * r
                + _buttonWidth * Accordion.buttonMarginX / 2
            )
            for button in row:

                if all(
                    [
                        JazzNote.noteNameFlats[JazzNote.distanceFromC(n)]
                        in _changeNotes
                        for n in button
                    ]
                ):
                    # print('button{} x{} y{}'.format(button, _buttonX, _buttonY))
                    # print('button {} key{} FCircle{}'.format(button,key,[JazzNote.distanceFromC(i) - JazzNote.distanceFromC(key) for i in button]))
                    _changeButton: Change = Change.makeFromSet(
                        [
                            (JazzNote.distanceFromC(i) - JazzNote.distanceFromC(key))
                            % 12
                            for i in button
                        ]
                    )
                    _changeButton = _changeButton.sortBySemitonePosition()
                    # print('changeButton',_changeButton)
                    if indicateNotesWithCircle:
                        _draw.ellipse(
                            (
                                _buttonX,
                                _buttonY,
                                _buttonX + _buttonWidth,
                                _buttonY + _buttonHeight,
                            ),
                            fill=Accordion.buttonIndicatedBGColour,
                            outline=Accordion.buttonIndicatedOutlineColour,
                        )

                    if indicateNotesWithFCircle:
                        Graphics.pasteFCircleOnIm(
                            im=_im,
                            xy=(
                                _buttonX,
                                _buttonY,
                                _buttonX + _buttonWidth,
                                _buttonY + _buttonHeight,
                            ),
                            change=_changeButton,
                            key=key,
                            greyScale=False,
                            invertColour=False,
                        )
                        # input('indicating.......')

                else:

                    _draw.ellipse(
                        (
                            _buttonX,
                            _buttonY,
                            _buttonX + _buttonWidth,
                            _buttonY + _buttonHeight,
                        ),
                        fill=Accordion.buttonBGColour,
                        outline=Accordion.buttonOutlineColour,
                    )
                _buttonX += _buttonWidth * (1 + Accordion.buttonMarginX)
            # _buttonX += _buttonWidth * (Accordion.rowOffsetByButtons)
            _buttonY += _buttonHeight * (1 + Accordion.buttonMarginY)
        _fullname = Accordion.getAccordionPath(
            change, key, greyScale, invertColour=invertColour
        )
        _pathname = Accordion.getAccordionPath(
            change, key, greyScale, invertColour=invertColour, returnName=False
        )
        if os.path.isdir(_pathname):
            pass
        else:
            Utility.makeDirectory(_pathname)
        print("saving {} to {}".format(_fullname, _pathname))
        _im.save(_fullname)
        Graphics.transposeColourSchemes(_im, _fullname, keyOrig=key)
        if showIm:
            _im.show()
        _im.save(_fullname)
        # Colour trranspose


from __future__ import annotations, print_function
from Key import Key
from Utility import Utility
from Book import Book
from JazzNote import JazzNote

import os
input = Utility.input
class NGramGraphic:
    """@classmethod
    def nGramPath(cls, rootKey: Key, noteset: [int], lines=6, invertedColours=False,
                  greyScale=False, filetype='pdf', ):
        if not lines in (3, 4, 6): raise ValueError('Invalid lines.')
        if lines == 3:
            nGramType = 'Trigram'
            subpage = Trigram.notesets[noteset]
        elif lines == 4:
            nGramType = 'Tetragram'
            subpage = Tetragram.notesets[noteset]
        elif lines == 6:
            nGramType = 'Hexagram'
            subpage = Hexagram.notesets[noteset]
        filename = nGramType + '_' + subpage
        path = os.path.join(Graphics.directory, nGramType, filename)

        print('here she be')
        input(path)

    @classmethod
    def renderNGrams(cls,
                     rootKeys: [Key] = [Key(n) for n in Key.allFlats],
                     filetypes=['pdf']):
        rootKeys = [Key(n) for n in rootKeys]
        print('making Hexagrams.')
        for noteset in Hexagram.notesets:
            hexagram = Hexagram.notesets[noteset]
            print('Hexagram', Hexagram['subpage'], 'in ')
            for key in rootKeys:
                print(key, end=', ')

            print()"""

    validTypes = [
        "Monogram",
        "Digram",
        "Trigram",
        "Tetragram",
        "Pentagram",
        "Hexagram",
        "Septagram",
        "Octagram",
        "Nantogram",
        "Decagram",
        "Monodecagram",
        "Duadecagram",
    ]
    '''# This is confusing, what can you do!? ;)
    Graphics.validDiagrams += validTypes'''
    aspectRatio = 1
    lineSpacing = 0.61803399
    breakSpacing = 0.23606797893228276  # small phi cubed
    # breakSpacing = .333333333333333333333333333333333333333333333333
    penThickness = 1
    lineColour = (0, 0, 0, 255)
    lineColourInvert = (255, 255, 255, 255)
    lineCountToDimensions = {
        "6": {"width": 11 / 12, "height": 1},
        "3": {"width": 1, "height": 1},
        "default": {"width": 1, "height": 1},
    }

    def __init__(self, lineCount=6, nGramType=None):
        """using nGramType == 'Trigram' will override using lineCount"""
        if nGramType == None:
            self.lineCount = lineCount
            self.nGramType = NGramGraphic.validTypes[int(lineCount) - 1]
        else:
            self.nGramType = nGramType
            self.lineCount = NGramGraphic.validTypes.index(nGramType) + 1
        self.book = Book(notesToUse=JazzNote.numberOctaveFlats[:lineCount])
        if lineCount != 6:
            warnings.warn(
                "Certain stuff has not been implemented for non-Hexagram ngrams yet."
            )
        if str(self.lineCount) in NGramGraphic.lineCountToDimensions.keys():
            self.canvasWidth = NGramGraphic.lineCountToDimensions[str(self.lineCount)][
                "width"
            ]
            self.canvasHeight = NGramGraphic.lineCountToDimensions[str(self.lineCount)][
                "height"
            ]
        else:
            self.canvasWidth = NGramGraphic.lineCountToDimensions["default"]["width"]
            self.canvasHeight = NGramGraphic.lineCountToDimensions["default"]["height"]

    def renderFiles(self, externalGraphicsPath=None):
        if externalGraphicsPath == None:
            from .Graphics import Graphics
            externalGraphicsPath = Graphics.externalGraphicsPath

        path = pyx.path
        deco = pyx.deco
        color = pyx.color
        style = pyx.style

        for change in self.book:
            semitones = change.getSemitones()
            for root in Key.allFlats:
                for greyScale in (False, True):
                    for invertColour in (False, True):
                        if not invertColour:
                            lineColour = NGramGraphic.lineColour
                            lineColourInv = NGramGraphic.lineColourInv
                        elif invertColour:
                            lineColour = NGramGraphic.lineColourInv
                            lineColourInv = NGramGraphic.lineColour
                        '''colours = Colour.getTransposedColours(
                            Key(root).distanceFromC(),
                            greyScale=greyScale,
                            invertColour=invertColour,
                        )'''
                        colours = Colour.getTransposedColours(
                            Key(root).distanceFromC(),neutralColours=True

                        )
                        filepath = self.getPath(
                            change=change,
                            key=root,
                            invertColour=invertColour,
                            greyScale=greyScale,
                            includeExtension=True,
                            externalGraphicsPath=externalGraphicsPath,
                        )
                        filedir = self.getPath(
                            change=change,
                            key=root,
                            invertColour=invertColour,
                            greyScale=greyScale,
                            includeFile=False,
                            externalGraphicsPath=externalGraphicsPath,
                        )

                        canvas = pyx.canvas.canvas()

                        canvasHeight = self.canvasHeight
                        canvasWidth = self.canvasWidth
                        lineHeight = canvasHeight * self.lineSpacing / self.lineCount
                        lineSpaceHeight = (
                            canvasHeight * (1 - self.lineSpacing) / (self.lineCount - 1)
                        )
                        # nput('the dimensions are these: width: {} height: {}'.format(canvasWidth,canvasHeight))

                        "input( 6*(lineSpaceHeight + lineHeight))"
                        for l in range(self.lineCount):
                            y0 = (lineHeight + lineSpaceHeight) * l
                            y1 = y0 + lineHeight
                            x0, x1, x2, x3 = (
                                0,
                                (1 - self.breakSpacing) * canvasWidth / 2,
                                (1 + self.breakSpacing) * canvasWidth / 2,
                                canvasWidth,
                            )
                            if l in semitones:
                                rects = (
                                    path.path(
                                        path.moveto(x0, y0),
                                        path.lineto(x0, y1),
                                        path.lineto(x3, y1),
                                        path.lineto(x3, y0),
                                        path.lineto(x0, y0),
                                        path.closepath(),
                                    ),
                                )
                            else:
                                rects = (
                                    path.path(
                                        path.moveto(x0, y0),
                                        path.lineto(x0, y1),
                                        path.lineto(x1, y1),
                                        path.lineto(x1, y0),
                                        path.lineto(x0, y0),
                                        path.closepath(),
                                    ),
                                    path.path(
                                        path.moveto(x2, y0),
                                        path.lineto(x2, y1),
                                        path.lineto(x3, y1),
                                        path.lineto(x3, y0),
                                        path.lineto(x2, y0),
                                        path.closepath(),
                                    ),
                                )

                            canvas.linewidth = 1
                            colour = colours[l]
                            """canvas.stroke(rect2,[
                                color.rgb(lineColour[0] / 255,lineColour[1] / 255,lineColour[2] / 255)])
                            canvas.fill(rect2,[
                                color.rgb(colour[0] / 255,colour[1] / 255,colour[2] / 255)])"""

                            for rect in rects:
                                if l in semitones:
                                    fillColour = colour
                                else:
                                    fillColour = lineColourInv
                                canvas.draw(
                                    rect,
                                    [
                                        style.linejoin.round,
                                        color.rgb(
                                            lineColour[0] / 255,
                                            lineColour[1] / 255,
                                            lineColour[2] / 255,
                                        ),
                                        deco.stroked(),
                                        deco.filled(
                                            [
                                                color.rgb(
                                                    fillColour[0] / 255,
                                                    fillColour[1] / 255,
                                                    fillColour[2] / 255,
                                                )
                                            ]
                                        ),
                                    ],
                                )

                        Utility.makeDirectory(filedir)
                        canvas.writetofile(filepath)
                        print(
                            "saved {} {} to file///:{}.".format(
                                self.nGramType,
                                change.getChangeNumber(combinatoricSize=self.lineCount),
                                filepath,
                            )
                        )

    def getPath(
        self,
        change: Change,
        key=Key("C"),
        invertColour=False,
        greyScale=False,
        includeGraphicsPath=True,
        externalGraphicsPath=False,
        includeFile=True,
        includeExtension=False,
        pathSeperator="\\",
    ):

        key = Key(key)
        nGramType = NGramGraphic.validTypes[self.lineCount - 1]
        from Graphics import Graphics
        if invertColour == False:
            if greyScale == False:
                path = Graphics.hexagramsDirectory
            else:
                path = Graphics.hexagramsBWDirectory
        else:
            if greyScale == False:
                path = Graphics.hexagramsInvDirectory
            else:
                path = Graphics.hexagramsBWInvDirectory
        if not includeGraphicsPath:
            path = "Hexagram" + pathSeperator
        path = os.path.join(path,key.getASCII())
        if includeFile:
            path = os.path.join(path,"Hexagram_" + str(
                change.getChangeNumber(
                    combinatoricSize=self.lineCount,
                )
            ))
            if includeExtension:
                path += ".pdf"
        path = path.replace("\\", pathSeperator).replace("/", pathSeperator)
        path = path.replace("Hexagram", nGramType)
        # input('in NGramGraphic.getPath path: {} includeExtension: {}'.format(path,includeExtension))
        if externalGraphicsPath:
            path = path.replace(
                Graphics.directory.replace("\\", pathSeperator).replace(
                    "/", pathSeperator
                ),
                Graphics.directoryExternal.replace("\\", pathSeperator).replace(
                    "/", pathSeperator
                ),
            )
        return path


from Unicode import Unicode
from Utility import Utility
input = Utility.input
class FontMap:
    "🐶🐺🐱🐭🐹🐰🐸🐯🐨🐻🐷🐽🐮🐗🐵🐒🐴🐑🐘🐼🐧🐦🐤🐥🐣🐔🐍🐢🐛🐝🐜🐿🐞🐌🐙🐚🐠🐟🐬🐳🐋🐄🐏🐀🐃🐅🐇🐉🐎🐐🐓🐕🐖🐁🐂🐲🐡🐊🐫🐪🐆🐈🐩🐾🦀🦁🦂🕷🦃🦄🦐🦑🦋🦍🦊🦌🦏🦇🦅🦆🦉🦎🦈🦓🦒🦔🦕🦖🦝🦙🦛🦘🦡🦢🦚🦜🦟🐻‍❄️🦭🐃🪱🦬🪳🦣🦤🐈‍⬛🦗🦫🪲🪰"
    "No, abacus treble clef  or plus minus unicode disable"

    latexFontStyles = [
        "BoldFont",
        "ItalicFont",
        "BoldItalicFont",
        "SlantedFont",
        "BoldSlantedFont",
        "SmallCapsFont",
        "UprightFont",
    ]
    fontFamilies = {
        "chinese": {
            "latexName": "cj",
            "fontName": "SimSun",
            "chars": ["。", "道德经", "易經", "易经", "易經", "太玄經"],
        },
        "symbol 1 alt": {
            "latexName": "symbZ",
            "fontName": "Symbola",
            "chars": [
                         "🐜","🎪", "🦀",'🐛','☯','❄',

                "𝄻",
                "𝄼",
                "𝄽",
                "𝄾",
                "🠥",
                "🠧",
                "➿",
                "🔮",

                "👐",
                "🍀",
                "🍂",
                "💐",
                "🏯",
                "🍁",
                "🌿",
                "❂",
                "🌴",
                "⚊",
                "⚋",
                "𝅜",
                "𝅘𝅥𝅯",
                "𝅘𝅥𝅰",
                "𝅘𝅥𝅱",
                "𝅘𝅥𝅲",

                "🐭",
                "🐘",
                "🐡",
                "🐍",
                "🐞",
                "🦀",  # '\U0001F600'🦀,🦀 #Crab🦀🦀🦀🦀🦀🦀🦀🦀🦀'
                "🐙",
                "🐌",
                "🙉",
                "⨍",
                "🐒",
                "🐫",
                "🐳",
                "🐰",
                "🌈",
                "🌱",
                "🌵",
                     ]
                     
        },

        "symbol 1": {
            "latexName": "symbA",
            #"fontName": "Symbola_monospacified_for_DejaVuSansMono",
            "fontName": "Symbola_monospacified_for_DejaVuSansMono",
            "chars": [

            ]
            + Unicode.romanNumeralCapitals
            + Unicode.romanNumeralLowercase,
        },
        "symbol black": {
            "latexName": "symbB",
            "fontName": "seguisym",
            "chars": [
                "⚬",

                "│",
                "┃",
                "╽",
                "╿",
                "╏",
                "║",
                "┆",
                "╵",
                "〡",
                "〢",
                "╹",
                "╻",
                "╷",
                "〣",
            ],
        },
        "music symbol": {
            "latexName": "symbM",
            "fontName": "MuseJazzText_9",
            "chars": ["♮", "♭", "♯", "𝄫", "𝄪", "𝅝", "𝅗𝅥", "♩", "♪", ""],
        },
        "obscure unicode": {
            "latexName": "symbC",
            "fontName": "unifont-11.0.02",
            "chars": ["🧮"],
        },
        "alt font 1": {"latexName": "altFA", "fontName": "sylfaen", "chars": []},
        "jazzA": {"latexName": "jazzA", "fontName": "JAZZTEXT", "chars": []},#JAZZTEXT, THWACK  PetalumaScript
        "jazzB": {"latexName": "jazzB", "fontName": "ADD-JAZZ", "chars": []},
        "british": {"latexName": "britA", "fontName": "rm-albion", "chars": []},
        "british embossed": {
            "latexName": "britB",
            "fontName": "GrafikText",
            "chars": [],
        },
        "celtic embossed": {
            "latexName": "britC",
            "fontName": "FrakturShadowed",
            "chars": [],
        },
        "block letter A": {
            "latexName": "boldA",
            "fontName": "AbrilFatface-Regular",
            "chars": [],
        },
        "block letter B": {
            "latexName": "boldB",
            "fontName": "GramophoneNF",
            "chars": [],
        },
        "extra bold": {
            "latexName": "boldC",
            "fontName": "texgyrebonum-bold",
            "chars": [],
        },
    }

    fontByWay = {
        "Jazz": {"latexName": "jazz", "fontName": "JAZZTEXT", "italic": "ADD-JAZZ"},
        "Carnatic": {
            "latexName": "carn",
            "fontName": "DorovarflfCarolus-axyg",
            "ItalicFont": "DorovarflfItalic-EaX8",
        },
        "Hexagram": {"latexName": "hexa", "fontName": "GramophoneNF"},
        "Word": {"latexName": "wayofword", "fontName": "MonkeyFingersNF"},
        #'Word': {'latexName': 'wayofword', 'fontName': 'DorovarflfCarolus-axyg','ItalicFont':'DorovarflfItalic-EaX8'},
        "Change Number": {
            "latexName": "changenumber",
            "fontName": "AbrilFatface-Regular",
        },
        "Scale Name": {"latexName": "scalename", "fontName": "AbrilFatface-Regular"},
        #'Change Number': {'latexName': 'changenumber', 'fontName': 'PastiRegular-mLXnm','ItalicFont':'PastiOblique-7B0wK'},
    }
    omitSymbols = [
        "",
        " ",
        "+/-",
        "+",
        "-",
        ",",
        ".",
        ":",
        "±",
        "№",
        "🧮",
        "°",
        "Δ",
        "a",
        "A",
        "b",
        "B",
        "c",
        "C",
        "d",
        "D",
        "e",
        "E",
        "f",
        "F",
        "g",
        "G",
        "h",
        "H",
        "i",
        "I",
        "j",
        "J",
        "k",
        "K",
        "l",
        "L",
        "m",
        "M",
        "n",
        "N",
        "o",
        "O",
        "p",
        "P",
        "q",
        "Q",
        "r",
        "R",
        "s",
        "S",
        "t",
        "T",
        "u",
        "U",
        "v",
        "V",
        "w",
        "W",
        "x",
        "X",
        "y",
        "Y",
        "z",
        "Z",
    ]

    cjkRanges = [
        {"from": ord("\u3300"), "to": ord("\u33ff")},  # compatibility ideographs
        {"from": ord("\ufe30"), "to": ord("\ufe4f")},  # compatibility ideographs
        {"from": ord("\uf900"), "to": ord("\ufaff")},  # compatibility ideographs
        {
            "from": ord("\U0002F800"),
            "to": ord("\U0002fa1f"),
        },  # compatibility ideographs
        {"from": ord("\u3040"), "to": ord("\u309f")},  # Japanese Hiragana
        {"from": ord("\u30a0"), "to": ord("\u30ff")},  # Japanese Katakana
        {"from": ord("\u2e80"), "to": ord("\u2eff")},  # cjk radicals supplement
        {"from": ord("\u4e00"), "to": ord("\u9fff")},
        {"from": ord("\u3400"), "to": ord("\u4dbf")},
        {"from": ord("\U00020000"), "to": ord("\U0002a6df")},
        {"from": ord("\U0002a700"), "to": ord("\U0002b73f")},
        {"from": ord("\U0002b740"), "to": ord("\U0002b81f")},
        {
            "from": ord("\U0002b820"),
            "to": ord("\U0002ceaf"),
        },  # included as of Unicode 8.0
    ]

    def is_cjk(char):
        return any(
            [range["from"] <= ord(char) <= range["to"] for range in FontMap.ranges]
        )

    @classmethod
    def makeFontMapTex(cls, absolutePathToFonts=False, saveEveryCharToFile=True):
        # https://texdoc.org/serve/fontspec/0
        _fontPathCommand = "\\FontPath"
        #_fontPath = "F:/wayofchange"
        _fontPath = os.getcwd()
        _tex = "\providecommand*{{{}}}{{{}}}\n".format(_fontPathCommand, _fontPath)
        _tex += """\RequirePackage{fontspec}
\RequirePackage{newunicodechar}
"""
        #input(_tex)
        _glyphs = []
        _glyphsChi = []
        _glyphsSym = []

        # for idx,i in enumerate({**Trigram.notesets, **Tetragram.notesets}):
        """for idx,i in enumerate({**Trigram.notesets, **Tetragram.notesets}):

            _nGram = [Trigram,Tetragram][idx]
            print(i, idx,_nGram,_nGram.notesets)
            for s in _nGram.notesets:
                for char in s:
                    _glyphs.append(_nGram.notesets[s]['chinese'])
       """
        for f in (FontMap.fontFamilies, FontMap.fontByWay):
            for i in f:
                _fam = f[i]
                _fontName = _fam["fontName"]
                _fontFiletype = None
                _fontCandidate = os.path.join(_fontPath, _fontName)
                for extension in ("otf", "ttf", "ttc"):
                    if _fontFiletype is None and os.path.exists(
                        _fontCandidate + "." + extension
                    ):
                        _fontFiletype = extension
                        _fontCandidate += "." + extension
                if _fontFiletype is None:
                    raise FileNotFoundError(
                        "The font does not exist at {}. You may not be able to render stuff right or other problems".format(_fontCandidate)
                    )
                if absolutePathToFonts:
                    _tex += (
                        "\\newfontfamily{\\"
                        + _fam["latexName"]
                        + "}{"
                        + _fontPath
                        + "/"
                        + _fam["fontName"]
                        + "."
                        + _fontFiletype
                        + "}%\n"
                    )
                else:
                    _tex += (
                        "\\newfontfamily{\\"
                        + _fam["latexName"]
                        + "}{"
                        + _fam["fontName"]
                        + "."
                        + _fontFiletype
                        + "}"
                    )
                    _fontVariants = {}
                    for k, variant in enumerate(FontMap.latexFontStyles):
                        if variant in _fam:
                            # input(FontMap.latexFontStyles[k])
                            # input(str(_fam[variant]))
                            _fontVariantFiletype = None
                            _fontCandidate = os.path.join(
                                _fontPath, _fontName
                            )
                            for extension in ("otf", "ttf", "ttc"):
                                if _fontVariantFiletype is None and os.path.exists(
                                    _fontCandidate + "." + extension
                                ):
                                    _fontVariantFiletype = extension
                                    _fontCandidate += "." + extension
                                    _fontVariants[variant] = {
                                        "fontName": _fontCandidate,
                                        "Extension": extension,
                                    }
                            if _fontVariantFiletype is None:
                                raise FileNotFoundError(
                                    "The font does not exist at {}".format(
                                        _fontCandidate
                                    )
                                )
                    # _tex += '[\n    Extension = .'+_fontFiletype+' ,'

                    if len(_fontVariants) >= 1:
                        _tex += "["
                        for k, variant in enumerate(_fontVariants):
                            _tex += (
                                "\n    "
                                + variant
                                + " = { "
                                + _fam[variant]
                                + "."
                                + _fontVariantFiletype
                                + " }"
                            )
                            if k < (len(_fontVariants) - 1):
                                _tex += ","
                        _tex += "\n]"
                    # _tex += '\n]'

                    _tex += "%\n"

        # \newfontfamily{\cj}{SimSun}

        # Get chinese in
        for i in Hexagram.chinese:
            for char in i:
                _glyphsChi.append(char)
        for s in Trigram.notesets:
            _chinese = Trigram.notesets[s]["chinese"]
            for char in _chinese:
                _glyphsChi.append(char)
        for s in Tetragram.semitones:
            _chinese = Tetragram.semitones[s]["chinese"]
            for char in _chinese:
                _glyphsChi.append(char)
        for s in Tetragram.ternaryList:
            _chinese = Tetragram.ternaryList[s]["chinese"]
            for char in _chinese:
                _glyphsChi.append(char)
        _glyphsChi.append(Unicode.chars["Hexagram Chinese"])
        _glyphsChi.append(Unicode.chars["Tetragram Chinese"])
        for char in Unicode.chars["IChing Chinese"]:
            _glyphsChi.append(char)
        for i in FontMap.fontFamilies["chinese"]["chars"]:
            for c in i:
                _glyphsChi.append(c)
        for i in _glyphsChi:
            if i not in FontMap.omitSymbols and i not in _glyphs:
                _tex += (
                    "\\newunicodechar {"
                    + i
                    + "}{{\\"
                    + FontMap.fontFamilies["chinese"]["latexName"]
                    + " "
                    + i
                    + "}}"
                )
                _tex += "%\n"
                _glyphs.append(i)

        for i in Unicode.chars:
            i = Unicode.chars[i].replace(" ", "").replace("\hfill", "")
            _fontFamily = FontMap.fontFamilies["symbol 1"]["latexName"]
            if i not in FontMap.omitSymbols and i not in _glyphs:
                _fontFamily = FontMap.fontFamilies["symbol 1"]["latexName"]
                for f in FontMap.fontFamilies:
                    if i in FontMap.fontFamilies[f]["chars"]:
                        _fontFamily = FontMap.fontFamilies[f]["latexName"]
                        break
                for c in i:
                    if c not in FontMap.omitSymbols and c not in _glyphs:
                        _hackStrEnd = ''
                        _hackStrBeginning = ''
                        if i == Unicode.chars['Sharp']:
                            #https://tex.stackexchange.com/questions/57788/understanding-ifnextchar
                            _hackStrEnd =  ' \\kern-.42ex'
                            _hackStrBeginning =''
                            #input('abddfbadbfbdf')
                        _tex += (
                            "\\newunicodechar {"
                            + c
                            + "}{"+_hackStrBeginning+"{\\"
                            + _fontFamily
                            + " "
                            + c
                            + "}}{}}}".format(_hackStrEnd)
                            + "%\n"
                        )
                        _glyphsSym.append(i)
                        _glyphs.append(i)
        for i in Zodiac.semitoneToZodiac:
            _fontFamily = FontMap.fontFamilies["symbol 1"]["latexName"]
            for c in i:
                if c not in FontMap.omitSymbols and c not in _glyphs:
                    _tex += (
                        "\\newunicodechar {"
                        + c
                        + "}{{\\"
                        + _fontFamily
                        + " "
                        + c
                        + "}}"
                        + "%\n"
                    )
                _glyphs.append(c)
        for i in Braille.semitonesFF:
            _fontFamily = FontMap.fontFamilies["symbol 1"]["latexName"]
            i = Braille.semitonesFF[i]
            # input(i)
            for c in i:
                if c not in FontMap.omitSymbols and c not in _glyphs:
                    _tex += (
                        "\\newunicodechar {"
                        + c
                        + "}{{\\"
                        + _fontFamily
                        + " "
                        + c
                        + "}}"
                        + "%\n"
                    )
                _glyphs.append(c)
        for i in (
            Hexagram.symbol[1:]
            + [Tetragram.semitones[t]["symbol"] for t in Tetragram.semitones.keys()]
            + [Trigram.notesets[t]["symbol"] for t in Trigram.notesets.keys()]
        ):
            _fontFamily = FontMap.fontFamilies["symbol black"]["latexName"]

            print(i,end=' ')
            for c in i:
                if c not in FontMap.omitSymbols and c not in _glyphs:
                    _tex += (
                        "\\newunicodechar {"
                        + c
                        + "}{{\\"
                        + _fontFamily
                        + " "
                        + c
                        + "}}"
                        + "%\n"
                    )
                _glyphs.append(c)

        for i in list(Change.infoWaySymbols):
            i = Change.infoWaySymbols[i].replace(" ", "").replace("\hfill", "")
            _fontFamily = FontMap.fontFamilies["symbol 1"]["latexName"]
            for f in FontMap.fontFamilies:
                if i in FontMap.fontFamilies[f]["chars"]:
                    _fontFamily = FontMap.fontFamilies[f]["latexName"]
                    break
            if i not in FontMap.omitSymbols and i not in _glyphs:
                for c in i:
                    if c not in FontMap.omitSymbols and c not in _glyphs:
                        _tex += (
                            "\\newunicodechar {"
                            + c
                            + "}{{\\"
                            + _fontFamily
                            + " "
                            + c
                            + "}}"
                            + "%\n"
                        )
                _glyphsSym.append(i)
                _glyphs.append(i)
        for family in FontMap.fontFamilies:
            for symbols in FontMap.fontFamilies[family]["chars"]:
                for char in symbols:
                    if char not in FontMap.omitSymbols and char not in _glyphs:
                        _tex += (
                            "\\newunicodechar {"
                            + char
                            + "}{{\\"
                            + FontMap.fontFamilies[family]["latexName"]
                            + " "
                            + char
                            + "}}"
                            + "%\n"
                        )
                        _glyphs.append(char)
        # check it out
        if len(set(_glyphs)) != len(_glyphs):
            _differenceInGlyphs = set([x for x in _glyphs if _glyphs.count(x) > 1])
            print("Shouldn't be duplicates in glyphs\n{}. Duplicates are: {}".format(
                    _glyphs, _differenceInGlyphs))
            raise TypeError(
                "Shouldn't be duplicates in glyphs\n{}. Duplicates are: {}".format(
                    _glyphs, _differenceInGlyphs
            )
            )
        '''
\directlua
{
 fonts.handlers.otf.addfeature
  {
    name = "ktest",
    type = "kern",
    data =
        {
            ["1"] = { 
                      ["1"] =  -200 ,
                      ["2"] =  -200 
                    },
        },
  }
 }'''
        _filename = Project.fontMapTexPath
        text_file = open(_filename, "w+", encoding="UTF-8")
        if _tex[-1] == "\n":
            _tex = _tex[0:-1]
        text_file.write(_tex)
        # text_file.write("%s" % _tex)
        text_file.close()
        print(_tex)

        print("glyphs " + " ".join(_glyphs))
        print("updated " + _filename)
        print("no of glyphs", len(_glyphs))
        print("updated the file: fontmap at " + _filename)
        if saveEveryCharToFile:
            _filename = _filename.replace('.tex','CharSample.tex')
            text_file = open(_filename, "w+", encoding="UTF-8")
            text_file.write(' '.join(_glyphs))
            text_file.close()
            print('wrote 1 of every glyph to {}'.format(_filename))

        return _tex

        # Hexagrams
        # Book symbols


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wayofchange']

package_data = \
{'': ['*'], 'wayofchange': ['.idea/*', '.idea/inspectionProfiles/*', 'fonts/*']}

install_requires = \
['os', 'pdf2image', 'pdfCropMargins', 'python-polylabel', 'pyx', 'tqdm']

setup_kwargs = {
    'name': 'wayofchange',
    'version': '0.0.7',
    'description': 'Python suite for music theory analysis and LaTeX outputs.',
    'long_description': 'Read me now',
    'author': 'Edrihan Levesque',
    'author_email': 'edrihan@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from Utility import Utility
from Change import Change
from Key import Key
from JazzNote import JazzNote
from Colour import Colour
import pyx,os,math
input = Utility.input

class Piano:
    height = 256
    fCircleVectorResolution = 64
    defaultVectorResolution = fCircleVectorResolution
    lineWidth = round(height / 64)
    lineWidthProportion = 1 / 64
    octaveWidthProportion = 2  # 1.8
    lineColour = (0, 0, 0, 255)
    lineColourInvert = (255, 255, 255, 255)
    indicationColour = (255, 255, 255, 255)
    indicationColourInvert = (255, 255, 255, 255)
    indicationLineColour = (0, 0, 0, 255)  # (128,128,128,255)
    indicationLineColourInvert = (0, 0, 0, 255)  # (128,128,128,255)
    canvasBGColour = (255, 255, 255, 0)
    blackKeyProportion = 0.61832  # 0.56
    evenlySpacedWhites = True
    if evenlySpacedWhites:
        CtoEWhites = 35
        FtoBWhites = 35
        CtoEBlacks = 21
        FtoBBlacks = 20
    elif not evenlySpacedWhites:
        # Evenly-spaced blacks
        CtoEWhites = 20
        FtoBWhites = 21
        CtoEBlacks = 12
        FtoBBlacks = 12

    def __init__(
        self,
        keyboardHeightPx=None,
        octaveWidthProportion=None,
        lineColour=None,
        lineWidthProportion=1 / 64,
        canvasBGColour=None,
        blackKeyProportion=None,  # 0.61832
        evenlySpacedWhites=None,
    ):

        if keyboardHeightPx is None:
            self.height = Piano.height
        else:
            self.height = keyboardHeightPx
        if octaveWidthProportion is None:
            self.octaveWidthProportion = Piano.octaveWidthProportion
        else:
            self.octaveWidthProportion = octaveWidthProportion
        if lineColour is None:
            self.lineColour = Piano.lineColour
        else:
            self.lineColour = lineColour
        if canvasBGColour is None:
            self.canvasBGColour = Piano.canvasBGColour
        else:
            self.canvasBGColour = canvasBGColour
        if blackKeyProportion is None:
            self.blackKeyProportion = Piano.blackKeyProportion
        else:
            self.blackKeyProportion = blackKeyProportion
        if evenlySpacedWhites is None:
            self.evenlySpacedWhites = Piano.evenlySpacedWhites
        else:
            self.evenlySpacedWhites = evenlySpacedWhites
        if lineWidthProportion is None:
            self.lineWidthProportion = Piano.lineWidthProportion
        else:
            self.lineWidthProportion = lineWidthProportion
        if evenlySpacedWhites:
            self.CtoEWhites = 35
            self.FtoBWhites = 35
            self.CtoEBlacks = 21
            self.FtoBBlacks = 20
        elif not evenlySpacedWhites:
            # Evenly-spaced blacks
            self.CtoEWhites = 20
            self.FtoBWhites = 21
            self.CtoEBlacks = 12
            self.FtoBBlacks = 12
        _octaveWidthBefore = 3 * self.CtoEWhites + 4 * self.FtoBWhites
        self.octaveWidth = round(self.height) * self.octaveWidthProportion
        _resProportion = self.octaveWidth / _octaveWidthBefore
        self.CtoEWhites = round(self.CtoEWhites * _resProportion)
        self.FtoBWhites = round(self.FtoBWhites * _resProportion)
        self.CtoEBlacks = round(self.CtoEBlacks * _resProportion)
        self.FtoBBlacks = round(self.FtoBBlacks * _resProportion)
        self.blackKeyLength = round(self.height * self.blackKeyProportion)

        self.lineWidth = round(self.height * self.lineWidthProportion)

    def insertGraphic(
        self,
        change: Change,
        key: str,
        resolution=None,
        greyScale=None,
        filetype="pdf",
        includeFiletypeExtension=False,
    ):
        if resolution is None:
            if filetype == "png":
                resolution = Piano.height
            else:
                resolution = Piano.defaultVectorResolution
        if greyScale is None:
            greyScale = not Colour.makeGraphicsColoured
        _filename = Piano.getPianoPath(
            change,
            key,
            resolution,
            greyScale=greyScale,
            includeGraphicsPath=not Latex.useRelativeGraphicsPath,
            filetype=filetype,
            includeFiletypeExtension=includeFiletypeExtension,
        )
        latexCommand = (
            Latex.commandStrings["Small Piano Command Start"]
            if filetype != "svg"
            else Latex.commandStrings["Small Piano Command Start Svg"]
        )
        return (
            latexCommand + _filename + Latex.commandStrings["Small Piano Command End"]
        )

    def makePianoFiles(
        self,
        changes: [Change],
        keys: [str],
        voicingType="Scale",
        startingKey=None,
        greyScale=False,
        alwaysFromCorF=False,
        invertColour=True,
        showDebug=False,
        indicateNotesWithDiagramType=False,
        makeNewIndicatorDiagrams=False,
        filetypes=["png"],
    ): #TODO: Make it support non fcircle. indicatorDiagramType does nothing currently

        # _polys = {}
        from Graphics import Graphics
        if changes == None:
            changes = range(-2048,2049)
        _useVector = any([f in Graphics.vectorFiletypes for f in filetypes])
        _useRaster = "png" in filetypes
        _startingKeyInitial = startingKey
        for key in keys:
            startingKey = _startingKeyInitial
            if alwaysFromCorF:
                if key in (
                    "C",
                    "Db",
                    "D",
                    "Eb",
                    "E",
                ):
                    startingKey = "C"
                    _totalKeys = 17
                else:
                    startingKey = "F"
                    _totalKeys = 19
            else:
                if startingKey == None:
                    startingKey = key
                else:
                    startingKey = key#raise ValueError('wtf {} {}'.format(startingKey,key))
                _totalKeys = 12
            if voicingType == "Thirds":
                _totalKeys += 12
            elif voicingType == 'Scale':
                pass
            else:
                raise ValueError('{} not a supported voicingType.'.format(voicingType))
            print(alwaysFromCorF,startingKey,_totalKeys)
            #input()
            # This is inefficient to have here
            _polys = self.getPianoKeyPolygons(
                startingKey=startingKey, totalKeys=_totalKeys
            )

            _pianoWidthPx = max([i for i in _polys[-1][::2]])
            fCircleCanvases = []
            fCircleCanvasesInv = []
            if indicateNotesWithDiagramType == False:
                pass
            else:
                for st in range(12):
                    #if True:
                    if any([st in Change.makeFromChangeNumber(c).getSemitones() for c in changes]):
                        print("rendering FCircle for piano")
                        fCircleCanvases.append(
                            Graphics.renderFCircle(
                                change=Change.makeFromSet([st]),
                                greyScale=greyScale,
                                invertColour=invertColour,
                                rootColourKey=key,
                                resolution=Piano.fCircleVectorResolution,
                                saveOnlyOneKey=True,
                                returnAsPyXCanvas=True,
                                filetypes=[],
                            )
                        )
                        fCircleCanvasesInv.append(
                            Graphics.renderFCircle(
                                change=Change.makeFromSet([st]),
                                greyScale=greyScale,
                                invertColour=not invertColour,
                                rootColourKey=key,
                                resolution=Piano.fCircleVectorResolution,
                                saveOnlyOneKey=True,
                                returnAsPyXCanvas=True,
                                filetypes=[],
                            )
                        )
                    else:
                        fCircleCanvases.append(None)
                        fCircleCanvasesInv.append(None)
            if type(changes[0]) == Change:
                changes = [c.getChangeNumber() for c in changes]
            for c in changes:
                
                change = Change.makeFromChangeNumber(c)
                if _useVector:
                    _c = pyx.canvas.canvas()

                if _useRaster:
                    _im = Image.new(
                        "RGBA",
                        (round(_pianoWidthPx), round(self.height)),
                        Piano.canvasBGColour,
                    )
                    _draw = ImageDraw.Draw(_im)

                self.drawPianoToIm(
                    im=_im if _useRaster else None,
                    canvas=_c if _useVector else None,
                    indicatorCanvasesSet=fCircleCanvases if (_useVector and indicateNotesWithDiagramType) else None,
                    indicatorCanvasesSetInv=fCircleCanvasesInv if (_useVector and indicateNotesWithDiagramType) else None,
                    change=change,
                    key=key,
                    octave=1,
                    startingKey=startingKey,
                    totalKeys=_totalKeys,
                    greyScale=greyScale,
                    invertColour=invertColour,
                    makeNewIndicatorDiagrams=makeNewIndicatorDiagrams,indicateNotesWithDiagramType=indicateNotesWithDiagramType,
                    voicingType=voicingType
                )

                for filetype in filetypes:
                    _fullname = Piano.getPianoPath(
                        change=change,
                        key=key,
                        resolution=self.defaultVectorResolution,
                        greyScale=greyScale,
                        invertColour=invertColour,
                        filetype=filetype,
                        voicingType=voicingType
                    )
                    _pathname = Piano.getPianoPath(
                        change=change,
                        key=key,
                        resolution=self.defaultVectorResolution,
                        greyScale=greyScale,
                        invertColour=invertColour,
                        includeFilename=False,
                        filetype=filetype,
                        voicingType=voicingType
                    )
                    if os.path.isdir(_pathname):
                        pass
                    else:
                        Utility.makeDirectory(_pathname)
                    if filetype == "png":
                        _im.save(_fullname)
                    else:
                        _finished = False
                        while not _finished:
                            try:
                                _c.writetofile(_fullname)
                                _finished = True
                            except PermissionError as e:
                                input(
                                    "{}\nTry to delete the file? If it does not work you may have to manually delete it. It may have gotten stuck. Windows thumbnail in exploror will do it if file is selected.".format(
                                        e
                                    )
                                )
                                if os.path.isfile(_fullname):
                                    os.remove(_fullname)
                                print("Removed the file.")

                    print(
                        "finished Making file:///"
                        + _fullname
                        + " . It used greyscale = {}, invertColour = {}".format(
                            greyScale, invertColour
                        )
                    )

    def drawPianoToIm(
        self,
        im,
        canvas: pyx.canvas.canvas,
        change: Change,
        key: Key,
        octave: int,
        startingKey: Key,
        totalKeys: int,
        greyScale,
        invertColour,
        voicingType: int = None,
        indicatorCanvasesSet: [pyx.canvas.canvas] = None,
        indicatorCanvasesSetInv: [pyx.canvas.canvas] = None,
        indicateNotesWithCircle=True,
        indicateNotesWithDiagramType=False,#"FCircle",
        showEachNoteOnce=True,
        makeBlackKeyIndicatorsAsBigAsWhites=True,
        showDebug=False,
        makeNewIndicatorDiagrams=False,
        useRadiusLines=True,
        fillInPartialKeys=True
    ):
        #assert type(change) == Change, 'fuck'
        
        if voicingType == 'Thirds':
            change = change.getVoicing(voicingType)
            #input(change)
        startingKey = Key(startingKey)
        if all([i == None for i in (im, canvas)]):
            raise ValueError("expecting im of type thingy or pyx.canvas")
        

        if canvas != None and len(change) > 0:
            if indicateNotesWithDiagramType not in (False,[]) and indicatorCanvasesSet == None:
                raise ValueError("Supply canvases for indicators.")
        canvas.layer("whiteKeys")
        canvas.layer("blackKeys")
        _useVector = canvas != None
        _useRaster = im != None
        _lineWidth = 3
        _startingDistFromC = Key(startingKey).distanceFromC()
        #input(key)
        _keyDistFromC = Key(key).distanceFromC()
        _colours = Colour.getTransposedColours(
            _startingDistFromC,
            greyScale=greyScale,
            invertColour=invertColour,
            pianoKeys=True,
        )

        _polys = self.getPianoKeyPolygons(startingKey=startingKey, totalKeys=totalKeys)
        _notes = [
            JazzNote.noteNameFlats[(i + _keyDistFromC) % 12]
            for i in change.getSemitones()
        ]
        # input('rah rah_polys {} \nlen \n{} \nkey {} \ndist from c{} \nnotes {}'.format(_polys,len(_polys),key,_keyDistFromC,_notes))
        if _useRaster:
            _draw = ImageDraw.Draw(im)
        if _useVector:
            _canvasScale = 1 / 48
            _lineWidthVec = pyx.style.linewidth(_canvasScale * 10)
            _lineWidthIndicatorCircleVec = pyx.style.linewidth(_canvasScale * 10 * 1)
            _trafo = pyx.trafo.scale(sx=_canvasScale, sy=-_canvasScale, x=0, y=0)
        _notesIndicated = 0
        _keysToIndicate = [
            i + _keyDistFromC - _startingDistFromC for i in change.getSemitones()
        ]
        _indicatorWhiteBoxes = []
        _indicatorBlackBoxes = []
        if not invertColour:
            _lineColour = Piano.lineColour
            _lineColourInvert = Piano.lineColourInvert
            _indicationColour = Piano.indicationColour
            _indicationColourInvert = Piano.indicationColourInvert
            _indicationLineColour = Piano.indicationLineColour
            _indicationLineColourInvert = Piano.indicationLineColourInvert
        else:
            _lineColour = Piano.lineColourInvert
            _lineColourInvert = Piano.lineColour
            _indicationColour = Piano.indicationColourInvert
            _indicationColourInvert = Piano.indicationColour
            _indicationLineColour = Piano.indicationLineColourInvert
            _indicationLineColourInvert = Piano.indicationLineColour
        _lineColourVec = pyx.color.rgb(
            _lineColour[0] / 255, _lineColour[1] / 255, _lineColour[2] / 255
        )
        _lineColourInvertVec = pyx.color.rgb(
            _lineColourInvert[0] / 255,
            _lineColourInvert[1] / 255,
            _lineColourInvert[2] / 255,
        )
        _indicationColourVec = pyx.color.rgb(
            _indicationColour[0] / 255,
            _indicationColour[1] / 255,
            _indicationColour[2] / 255,
        )
        _indicationColourInvertVec = pyx.color.rgb(
            _indicationColourInvert[0] / 255,
            _indicationColourInvert[1] / 255,
            _indicationColourInvert[2] / 255,
        )

        _indicationLineColourVec = pyx.color.rgb(
            _indicationLineColour[0] / 255,
            _indicationLineColour[1] / 255,
            _indicationLineColour[2] / 255,
        )
        _indicationLineColourInvertVec = pyx.color.rgb(
            _indicationLineColourInvert[0] / 255,
            _indicationLineColourInvert[1] / 255,
            _indicationLineColourInvert[2] / 255,
        )

        if fillInPartialKeys:
            def _partialKeyPath():
                polyPath = pyx.path.path()
                polyPath.append(pyx.path.moveto(_L, _T))
                polyPath.append(pyx.path.lineto(_L, _B))
                polyPath.append(pyx.path.lineto(_R, _B))
                polyPath.append(pyx.path.lineto(_R, _T))
                polyPath.append(pyx.path.closepath())
                return polyPath
            def _drawPartialWhiteKey():
                polyPath = _partialKeyPath()
                layCanvas = canvas.layer(
                    "whiteKeys",
                )

                layCanvas.stroke(
                    polyPath,
                    [
                        _lineWidthVec, _trafo, pyx.deco.stroked([_lineColourVec]),
                        pyx.deco.filled([_lineColourInvertVec]),
                    ],
                )
                return polyPath
            def _drawPartialBlackKey():
                polyPath = _partialKeyPath()
                layCanvas = canvas.layer(
                    "blackKeys",
                )

                layCanvas.stroke(
                    polyPath,
                    [
                        _lineWidthVec, _trafo, pyx.deco.stroked([_lineColourVec]),
                        pyx.deco.filled([_lineColourVec]),
                    ],
                )
                return polyPath

            # Right white key part
            if (startingKey + totalKeys - 1).isBlack():
                _L = max(_polys[-2][::2])
                _T = 0
                _B = max(_polys[totalKeys-2][1::2])
                if makeBlackKeyIndicatorsAsBigAsWhites == True:
                    _whitePoly = self.getPianoKeyPolygons(startingKey + totalKeys,1)[0]
                    _whiteW = max([i for i in _whitePoly[::2]]) - min(
                        [i for i in _whitePoly[::2]]
                    )
                    _R = (min(_polys[-1][::2]) + max(_polys[-1][::2])) / 2 + _whiteW /2
                else:
                    _R = max(_polys[totalKeys-2][::2])
                _drawPartialWhiteKey()
            # Left white key part
            if startingKey.isBlack():
                if makeBlackKeyIndicatorsAsBigAsWhites == True:
                    _whitePoly = _polys[1]
                    _whiteW = max([i for i in _whitePoly[::2]]) - min(
                        [i for i in _whitePoly[::2]]
                    )
                    _L = (min(_polys[0][::2]) + max(_polys[0][::2])) / 2 - _whiteW /2
                else:
                    _L = min(_polys[0][::2])

                #_L = -self.getPianoKeyPolygons(startingKey=startingKey-1, totalKeys=2)[1][::2]

                _R = min(_polys[1][::2])
                _T = 0
                _B = max(_polys[1][1::2])
                _drawPartialWhiteKey()
            # Right black key part
            if (startingKey + totalKeys ).isBlack():
                poly = _polys[-1]
                _L = Utility.secondBiggest(poly[::2])
                _R = max(poly[::2])
                _T = 0
                _B = Utility.secondBiggest(poly[1::2])
                _drawPartialBlackKey()
            #Left black key part
            if (startingKey - 1).isBlack():
                poly = self.getPianoKeyPolygons(startingKey - 1,1)[0]
                _T = 0
                _B = max(poly[1::2])
                _L = min(_polys[0][::2])
                _R = Utility.secondSmallest(_polys[0][::2])
                #input(_lineWidthVec)
                _drawPartialBlackKey()

        for idx, poly in enumerate(_polys):
            _note = JazzNote.noteNameFlats[(idx + _startingDistFromC) % 12]
            _note = Key(_note)

            # input(_note)

            ###DRAW KEYS
            if _useRaster:
                _draw.polygon(
                    poly,
                    fill=_colours[idx % 12],
                    outline=_lineColour,
                )
                Graphics.drawFatPolygonToIm(
                    im=im,
                    poly=poly,
                    outline=_lineColour,
                    width=_lineWidth,
                )
            if _useVector:

                polyPath = pyx.path.path()
                polyPath.append(pyx.path.moveto(poly[0], poly[1]))
                for p in range(2, len(poly), 2):
                    polyPath.append(pyx.path.lineto(poly[p], poly[p + 1]))

                polyPath.append(pyx.path.closepath())
                # input(_lineColour)

                # input('{} is whyte? {}'.format(_note,Key(_note).isWhite()))
                layCanvas = (
                    canvas.layer("whiteKeys")
                    if Key(_note).isWhite()
                    else canvas.layer(
                        "blackKeys",
                    )
                )

                layCanvas.stroke(
                    polyPath,
                    [
                        _lineWidthVec,
                        _trafo,
                        pyx.deco.stroked(
                            [_lineColourVec]
                            if Key(_note).isWhite()
                            else [_lineColourVec]
                        ),
                        pyx.deco.filled(
                            [_lineColourInvertVec]
                            if Key(_note).isWhite()
                            else [_lineColourVec]
                        ),
                    ],
                )

            if showDebug:
                print("_notes {} notes indicated {}".format(_notes, _notesIndicated))
            if (
                indicateNotesWithCircle
                and len(_notes) > 0
                and Key(_note) == Key(_notes[_notesIndicated % len(_notes)])
                and _notesIndicated < len(change)
            ):
                # print(len(change),'dfghdfghy')
                # input('this note {} is in {}'.format(_note, _notes))
                _iL = min([i for i in poly[::2]])
                _iR = max([i for i in poly[::2]])
                _iWidth = _iR - _iL
                _iB = max([i for i in poly[1::2]])
                _iT = _iB - _iWidth
                
                if _note.isWhite():
                    _indicatorWhiteBoxes.append(
                        {"L": _iL, "R": _iR, "T": _iT, "B": _iB, "N": _notesIndicated}
                    )
                    _notesIndicated += 1
                else:  # black note
                    if makeBlackKeyIndicatorsAsBigAsWhites == True:
                        if idx == 0:
                            _whitePoly = _polys[idx + 1]
                        else:
                            _whitePoly = _polys[idx - 1]
                        _whiteW = max([i for i in _whitePoly[::2]]) - min(
                            [i for i in _whitePoly[::2]]
                        )
                        _whiteDiff = (_whiteW - _iWidth) / 2
                        _iL -= _whiteDiff
                        _iR += _whiteDiff
                        _iT -= 2 * _whiteDiff
                        _iB += 0
                        _iWidth = _iR - _iL

                    _indicatorBlackBoxes.append(
                        {"L": _iL, "R": _iR, "T": _iT, "B": _iB, "N": _notesIndicated}
                    )
                    _notesIndicated += 1

        for indicators in [_indicatorBlackBoxes, _indicatorWhiteBoxes]:
            for indicator in indicators:
                _iL = indicator["L"]
                _iR = indicator["R"]
                _iT = indicator["T"]
                _iB = indicator["B"]
                _iN = indicator['N']
                #input(_iNote)
                _iWidth = _iR - _iL
                if indicators == _indicatorBlackBoxes:
                    _indicationColour = _indicationColour
                    _indicationLineColour = _indicationLineColour
                    _indicationColourVec = _indicationColourVec
                    _indicationLineColourVec = _indicationLineColourVec
                    _indicatorCanvasesSet = indicatorCanvasesSet

                else:
                    _indicationColour = _indicationColourInvert
                    _indicationLineColour = _indicationLineColourInvert
                    _indicationColourVec = _indicationColourInvertVec
                    _indicationLineColourVec = _indicationLineColourVec
                    _indicatorCanvasesSet = indicatorCanvasesSet  # Inv
                #input(indicateNotesWithDiagramType)
                if indicateNotesWithDiagramType in ([],False,None):
                    _indicationColour = Colour.getTransposedColours(
                        colourTranspose=Key(key).distanceFromC(),
                        greyScale=greyScale,neutralColours=True)[change[_iN].semitonesFromOne()]
                    #input(_indicationColour)
                    #input([i/255 for i in _indicationColour])
                    _indicationColourVec = pyx.color.rgb(_indicationColour[0]/255,
                                        _indicationColour[1]/255,
                                        _indicationColour[2]/255)
                    #input('indicator: {} _indicationColourVec: {}'.format(indicator, _indicationColour))

                    
                    
                
                if _useRaster:
                    _draw.ellipse(
                        [_iL, _iT, _iR, _iB],
                        fill=_indicationColour,
                        outline=_indicationLineColour,
                        width=_lineWidth,
                    )
                if _useVector:
                    # canvas.draw(pyx.path.circle((_iL+_iR)/2,(_iT+_iB)/2,_iWidth),
                    #            [_lineWidthVec,_trafo])

                    #print('iL iR',_iL,_iR)
                    _circleCanvas = pyx.canvas.canvas()
                    _circleCanvas.stroke(
                        pyx.path.circle((_iL + _iR) / 2, (_iT + _iB) / 2, _iWidth / 2),
                        [
                            _lineWidthIndicatorCircleVec,
                            _trafo,
                            pyx.deco.stroked(
                                [_lineColourVec]
                            ),
                            pyx.deco.filled(
                                [_indicationColourVec]

                            ),
                        ],
                    )
                    if useRadiusLines:
                        _noteChange = Change(change[_iN])
                        print(_iN)
                        #input(_noteChange)
                        _rad = abs(_iR - ((_iR + _iL) / 2))

                        _circleCanvas.stroke(pyx.path.line(
                            (_iR + _iL) / 2,
                            (_iT + _iB) / 2,
                            (_iR + _iL) / 2 + _rad * math.cos(
                                _noteChange.getAngles(returnDegrees=False, goQuarterTurnAnticlockwise=True)[0]),
                            (_iT + _iB) / 2 + _rad * math.sin(
                                _noteChange.getAngles(returnDegrees=False, goQuarterTurnAnticlockwise=True)[0]),
                        ),
                            [_lineWidthIndicatorCircleVec, pyx.style.linecap.round,
                             _trafo, _lineColourVec])
                    #print('iwidthvec :',_circleCanvas.bbox().right_pt() - _circleCanvas.bbox().left_pt())
                    _iWidthVec = (
                        _circleCanvas.bbox().right_pt() - _circleCanvas.bbox().left_pt()
                    )
                    # input(_iWidthVec)
                    canvas.insert(_circleCanvas)

                if indicateNotesWithDiagramType not in (None, [], False):

                    # _DiagramColourKey = key#JazzNote.noteNameFlats[(JazzNote.noteNameFlats.index(_note) )%12]# + _startingDistFromC

                    _changeToIndicate = Change(change.notes[indicator["N"]])
                    _distFromOne = _changeToIndicate[0].semitonesFromOne()
                    if _useVector:

                        _iCanvas = pyx.canvas.canvas()
                        _canvasWidth = (
                            canvas.bbox().left_pt() - canvas.bbox().right_pt()
                        )
                        # input(Key.allFlats[startingKey.distanceFromC()])

                        # _iTargetVectorWidth = _canvasWidth / _totalWhiteKeys

                        _iTranslateTrafo = pyx.trafo.translate(_iL, _iT)
                        _iCanvas.insert(
                            indicatorCanvasesSet[_distFromOne],
                            [_iTranslateTrafo,_trafo],
                        )
                        _iVectorWidth = (
                            _iCanvas.bbox().top_pt() - _iCanvas.bbox().bottom_pt()
                        )
                        _iScale = _iWidthVec / _iVectorWidth
                        #_iScale = _iVectorWidth/_iWidthVec 
                        _iScaleTrafo = pyx.trafo.scale(abs(_iScale), -abs(_iScale))
                        _iScaledCanvas = pyx.canvas.canvas()
                        _iScaledCanvas.insert(
                            _indicatorCanvasesSet[_distFromOne],
                            [
                                _iScaleTrafo,
                                _iTranslateTrafo,
                                _trafo
                            ],
                        )

                        # print(len(indicators),_iScaledCanvas.bbox().right_pt()-_iScaledCanvas.bbox().left_pt(),'is width')
                        if 0:

                            input(
                                "_iCanvas: {}, {} width: {}\ncanvas: {}, {} {} \n_iL: {}, _iR: {}".format(
                                    _iCanvas.bbox().top_pt(),
                                    _iCanvas.bbox().bottom_pt(),
                                    _iVectorWidth,
                                    canvas.bbox().top_pt(),
                                    canvas.bbox().bottom_pt(),
                                    _canvasWidth,
                                    _iL,
                                    _iR,
                                )
                            )

                        canvas = canvas.insert(_iScaledCanvas)
                        #canvas = canvas.insert(_iScaledCanvas,[_trafo])

                        """canvas.stroke(pyx.path.line(_iL, _iB,_iR, _iB) +
                                 pyx.path.line(_iR, _iB, _iR, _iT) +
                                 pyx.path.line(_iR, _iT, _iL, _iT) +
                                 pyx.path.line(_iL, _iT, _iL, _iB),[_trafo])"""
                        # print(change)
                        # input(_changeToIndicate)

                    if _useRaster:
                        _diagramPath = Graphics.getDiagramPath(
                            _changeToIndicate,
                            _FCircleColourKey,
                            int(_iWidth),
                            includeFilename=True,
                            greyScale=greyScale,
                            filetype="png",
                            diagramType=indicateNotesWithDiagramType,
                        )

                        if (not makeNewIndicatorDiagrams) and Graphics.seeIfDiagramExists(
                            _changeToIndicate,
                            _FCircleColourKey,
                            int(_iWidth),
                            greyScale=greyScale,
                            invertColour=invertColour,
                            impatientSkip=False,
                            filetype="png",
                            diagramType=indicateNotesWithDiagramType,
                        ):
                            pass
                        else:
                            # raise TypeError('{} Does not exist.'.format(_FCirclePath))
                            if makeNewDiagrams:
                                if diagramType == "FCircle":
                                    Graphics.renderFCircle(
                                        change=_changeToIndicate,
                                        resolution=int(_iWidth),
                                        rootColourKey="C",
                                        saveOnlyOneKey=False,
                                        filetypes=["png"],
                                    )
                                else:
                                    raise NotImplementedError(
                                        "change renderFCircle to Graphics.renderDiagram() once you make this"
                                    )
                            # raise TypeError('That FCircle don\'t exist: {}'.format(_FCirclePath))
                        _diagramFile = Image.open(_diagramPath).convert("RGBA")

                        im.paste(_diagramFile, (int(_iL), int(_iT)), _diagramFile)
                        # im.alpha_composite(_FCircleFile, dest=(int(_iL), int(_iT)), source=(0, 0))

                        # print('circle width'+str(_iWidth))
                        # print('just drew '+_FCirclePath+' '+' '.join([str(i) for i in [_iL,_iT,_iR,_iB]]))
        if _useRaster:
            # bottom line
            _draw.line(
                (0, self.height - _lineWidth / 2, 0, self.height - _lineWidth / 2),
                fill=_lineColour,
                width=_lineWidth,
            )
            # right line
            _draw.line(
                (
                    im.size[0] - _lineWidth / 2,
                    0,
                    im.size[0] - _lineWidth / 2,
                    self.height,
                ),
                fill=_lineColour,
                width=_lineWidth,
            )
            if showDebug:
                im.show()
            im = im.convert()
        returnDict = {}
        if _useRaster:
            returnDict["im"] = im
        if _useVector:
            returnDict["canvas"] = canvas
        return returnDict

    def getPianoKeyPolygons(
        self,
        startingKey: str,
        totalKeys: int,
    ) -> []:
        _keyPolys = []
        for keyname in (startingKey,):
            if keyname not in JazzNote.noteNameFlats:
                
                startingKey = Key(startingKey).inAllFlats().note
                #except:
                 #   raise TypeError("{} not in ".format(startingKey, JazzNote.noteNameFlats))
        # First we'll make it in C then we'll move it over
        # First lets make the first octave  of whites
        # C
        _Ckey = [
            0,
            0,
            0,
            self.height,
            self.CtoEWhites,
            self.height,
            self.CtoEWhites,
            self.blackKeyLength,
            self.CtoEBlacks,
            self.blackKeyLength,
            self.CtoEBlacks,
            0,
            0,
            0,
        ]
        # D
        _whiteL = self.CtoEWhites
        _whiteR = _whiteL + self.CtoEWhites
        _blackL = self.CtoEBlacks * 2
        _blackR = self.CtoEBlacks * 3

        _Dkey = [
            _blackL,
            0,
            _blackL,
            self.blackKeyLength,
            _whiteL,
            self.blackKeyLength,
            _whiteL,
            self.height,
            _whiteR,
            self.height,
            _whiteR,
            self.blackKeyLength,
            _blackR,
            self.blackKeyLength,
            _blackR,
            0,
            _blackL,
            0,
        ]
        # E
        _whiteL += self.CtoEWhites
        _whiteR = _whiteL + self.CtoEWhites
        _blackL += self.CtoEBlacks * 2
        # _blackR = _blackL + self.CtoEBlacks
        _Ekey = [
            _blackL,
            0,
            _blackL,
            self.blackKeyLength,
            _whiteL,
            self.blackKeyLength,
            _whiteL,
            self.height,
            _whiteR,
            self.height,
            _whiteR,
            0,
            _blackL,
            0,
        ]
        # F
        _whiteL += self.CtoEWhites
        _whiteR = _whiteL + self.FtoBWhites
        _blackL += self.CtoEBlacks * 2  # + self.FtoEBlacks
        # _blackR = _blackL + self.CtoEBlacks
        _Fkey = [
            _whiteL,
            0,
            _whiteL,
            self.height,
            _whiteR,
            self.height,
            _whiteR,
            self.blackKeyLength,
            _blackL,
            self.blackKeyLength,
            _blackL,
            0,
            _whiteL,
            0,
        ]
        # G
        _whiteL += self.FtoBWhites
        _whiteR = _whiteL + self.FtoBWhites
        _blackL += self.FtoBBlacks
        _blackR = _blackL + self.FtoBBlacks
        _Gkey = [
            _blackL,
            0,
            _blackL,
            self.blackKeyLength,
            _whiteL,
            self.blackKeyLength,
            _whiteL,
            self.height,
            _whiteR,
            self.height,
            _whiteR,
            self.blackKeyLength,
            _blackR,
            self.blackKeyLength,
            _blackR,
            0,
            _blackL,
            0,
        ]
        # A
        _whiteL += self.FtoBWhites
        _whiteR = _whiteL + self.FtoBWhites
        _blackL += self.FtoBBlacks * 2
        _blackR = _blackL + self.FtoBBlacks
        _Akey = [
            _blackL,
            0,
            _blackL,
            self.blackKeyLength,
            _whiteL,
            self.blackKeyLength,
            _whiteL,
            self.height,
            _whiteR,
            self.height,
            _whiteR,
            self.blackKeyLength,
            _blackR,
            self.blackKeyLength,
            _blackR,
            0,
            _blackL,
            0,
        ]
        # B
        _whiteL += self.FtoBWhites
        _whiteR = _whiteL + self.FtoBWhites
        _blackL += self.FtoBBlacks * 2
        # _blackR = _blackL + self.CtoEBlacks
        _Bkey = [
            _blackL,
            0,
            _blackL,
            self.blackKeyLength,
            _whiteL,
            self.blackKeyLength,
            _whiteL,
            self.height,
            _whiteR,
            self.height,
            _whiteR,
            0,
            _blackL,
            0,
        ]
        # Next make black keys
        # Db
        _blackL = self.CtoEBlacks
        _blackR = self.CtoEBlacks * 2
        _Dbkey = [
            _blackL,
            0,
            _blackL,
            self.blackKeyLength,
            _blackR,
            self.blackKeyLength,
            _blackR,
            0,
            _blackL,
            0,
        ]
        # Eb
        _blackL = self.CtoEBlacks * 3
        _blackR = self.CtoEBlacks * 4
        _Ebkey = [
            _blackL,
            0,
            _blackL,
            self.blackKeyLength,
            _blackR,
            self.blackKeyLength,
            _blackR,
            0,
            _blackL,
            0,
        ]
        # Gb
        _blackL = self.CtoEBlacks * 5 + self.FtoBBlacks
        _blackR = _blackL + self.FtoBBlacks
        _Gbkey = [
            _blackL,
            0,
            _blackL,
            self.blackKeyLength,
            _blackR,
            self.blackKeyLength,
            _blackR,
            0,
            _blackL,
            0,
        ]

        # Ab
        _blackL += self.FtoBBlacks * 2
        _blackR = _blackL + self.FtoBBlacks
        _Abkey = [
            _blackL,
            0,
            _blackL,
            self.blackKeyLength,
            _blackR,
            self.blackKeyLength,
            _blackR,
            0,
            _blackL,
            0,
        ]
        # Bb
        _blackL += self.FtoBBlacks * 2
        _blackR = _blackL + self.FtoBBlacks
        _Bbkey = [
            _blackL,
            0,
            _blackL,
            self.blackKeyLength,
            _blackR,
            self.blackKeyLength,
            _blackR,
            0,
            _blackL,
            0,
        ]

        # for key in (_Fkey,_Gbkey,):
        for idx, key in enumerate(
            (
                _Ckey,
                _Dbkey,
                _Dkey,
                _Ebkey,
                _Ekey,
                _Fkey,
                _Gbkey,
                _Gkey,
                _Abkey,
                _Akey,
                _Bbkey,
                _Bkey,
            )
        ):
            _keyName = JazzNote.noteNameFlats[idx]
            _keyPolys.append(key)
        _keyTransposition = JazzNote.noteNameFlats.index(startingKey)
        _xShift = min([x for x in _keyPolys[_keyTransposition][::2]])
        # print('xShift {}. startomgKey {}'.format(_xShift,startingKey))
        _firstPolys = _keyPolys[_keyTransposition:]
        _lastPolys = _keyPolys[0:_keyTransposition]
        for i in range(len(_firstPolys)):
            _poly = _firstPolys[i]
            for p, point in enumerate(_poly):
                if p % 2 == 0:
                    _poly[p] -= _xShift
            _firstPolys[i] = _poly
        for i in range(len(_lastPolys)):
            _poly = _lastPolys[i]
            for p, point in enumerate(_poly):
                if p % 2 == 0:
                    _poly[p] += -_xShift + self.octaveWidth
            _lastPolys[i] = _poly
        _keyPolys = _firstPolys + _lastPolys
        if totalKeys <= 12:
            return _keyPolys[:totalKeys]
        else:
            for nextKey in range(totalKeys - 12):
                _octavesUp = math.floor((nextKey + 12) / 12)
                _poly = _keyPolys[nextKey % 12][:]
                for p, point in enumerate(_poly):
                    if p % 2 == 0:
                        _poly[p] += self.octaveWidth * _octavesUp
                _keyPolys.append(_poly)
        # input('keypolys {}'.format(_keyPolys))
        return _keyPolys
        # return _firstPolys + _lastPolys
        # return _lastPolys

    @classmethod
    def getPianoPath(
        cls,
        change: Change,
        key: str,
        resolution: int = None,
        includeFilename=True,
        pathSeperator="/",
        greyScale=False,
        invertColour=False,
        filetype="pdf",
        includeFileExtension=True,
        includeGraphicsPath=True,
        externalGraphicsPath=False,
        voicingType=1
    ) -> str:

        if filetype == "pdf":
            if resolution == 256:
                raise ValueError("Piano only works in pdf at 64")
                print("256 not a valid resolution for Piano. Using 64")
        if resolution == None:
            resolution = Piano.defaultVectorResolution
        key = JazzNote.makeAlphabetNoteUseSingleFlats(key)


        _filename = "Piano"+("Thirds" if voicingType == 'Thirds' else '')+"_" + str(
            change.removeDuplicateNotes().getOctaveLimited().getChangeNumber(decorateChapter=False,)
        )
        from Graphics import Graphics
        if includeFileExtension:
            _filename += "." + filetype
        if greyScale:
            if invertColour:
                _path = Graphics.pianosBWInvDirectory
            else:
                _path = Graphics.pianosBWDirectory
        else:
            if invertColour:
                _path = Graphics.pianosInvDirectory
            else:
                _path = Graphics.pianosDirectory
        if externalGraphicsPath:
            _path = _path.replace(Graphics.directory, Graphics.directoryExternal)
        if not includeGraphicsPath:
            _path = "Piano"
        _path = os.path.join(_path,Graphics.getFiletypeResolutionFolderName(
            filetype=filetype, resolution=resolution
        ))
        _path = os.path.join(_path,key )

        # I did put key in filename
        if includeFilename:
            _str = os.path.join(_path , _filename)
        else:
            _str = _path
        _str = _str.replace("\\", pathSeperator)
        _str = _str.replace("/", pathSeperator)
        return _str